/*
 *  mvPCIe Driver - DMA implementation
 *  Copyright (c) 2019-2022 by MATRIX VISION GmbH <info@matrix-vision.de>
 *  Author: grayh (based on previous work by mrakk and others)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/uaccess.h>
#include <linux/of.h>
#include <linux/of_irq.h>
#include <linux/fs.h>
#include <linux/pci.h>
#include <linux/dma-direction.h>
#include <linux/version.h>

#include "mvdma.h"

static atomic_t image_count;

//#define GET_OUT_OF_JAIL_FREE_CARD // For development only.
//#define SHOW_HUNGER

#define _kB (1024LL)
#define _MB (1024LL * _kB)
#define _GB (1024LL * _MB)

static void mvdma_dma_callback(void *data, u32 img_size, u32 img_id)
{
    TmvUI_umem *um = (TmvUI_umem *) data;

    um->dma_state = DMA_DONE;
    um->img_size = img_size;
    um->img_id = img_id;
    wake_up(&um->wq);
}

static ssize_t mvdma_dma_to_user(TmvUI_umem * um)
{
    TmvUI_userhandle *uh = um->uh;
    struct mvdma_device *pdma = uh->pdma;

    dprintk_dma(MVDMA_DEBUG_QUEUE,
                "sl @ mem %d (%d entries), expect img_size 0x%x\n", um->nr, um->page_cnt, um->img_size);
    um->dma_state = DMA_RUNNING;

    return mvdma_request(uh->pdma, um->sl, um->page_cnt, mvdma_dma_callback, um);
    return 0;
}

static inline void mvdma_count_image(struct mvdma_device *pdma, u32 queue_nr)
{
    pdma->dma_queue[queue_nr]->img_id = image_count;
    atomic_inc(&image_count);
}

static int mvdma_clear_pool_for_queue(struct mvdma_device *pdma, unsigned queue_nr)
{
    int i;
    int block_count = pdma->dma_queue[queue_nr]->plrcnt;

    for (i = 0; i < block_count; i++) {
        struct page_list_req *plr = &pdma->dma_queue[queue_nr]->plr[i];

        //dprintk_dma(MVDMA_DEBUG_POOL, "%s: queue_nr = 0x%x, block_count = %d, i = %d\n", __FUNCTION__, queue_nr, block_count, i);
        if (free_page_list(pdma, plr)) {
            pr_err("%s: Error freeing page list for queue %d(%d)\n", __FUNCTION__, queue_nr, i);
            return -ENOMEM;
        }
    }
    return 0;
}

static inline void mvdma_deliver_queue(struct mvdma_device *pdma, int queue_nr)
{
    if (likely(pdma->dma_queue[queue_nr]->cb_user)) {
        if (!pdma->dma_queue[queue_nr]->valid_sgcnt)
            return;

        dprintk_dma(MVDMA_DEBUG_TASKLET, "%d done\n", queue_nr);
        dprintk_dma(MVDMA_DEBUG_MAPPING,
                    "%s: dma_unmap_sg pdma->dma_queue[%d]->sg (%px), valid_sgcnt=%d, sgcnt=%d\n",
                    __FUNCTION__, queue_nr,
                    pdma->dma_queue[queue_nr]->sg,
                    pdma->dma_queue[queue_nr]->valid_sgcnt, pdma->dma_queue[queue_nr]->sgcnt);
        /* the image buffer has been written by the device */
        dma_unmap_sg(pdma->dev, pdma->dma_queue[queue_nr]->sg, pdma->dma_queue[queue_nr]->sgcnt, DMA_FROM_DEVICE);
        mvdma_clear_pool_for_queue(pdma, queue_nr);
        pdma->dma_queue[queue_nr]->sgcnt = 0;
        pdma->dma_queue[queue_nr]->valid_sgcnt = 0;
        pdma->dma_queue[queue_nr]->cb_user(pdma->dma_queue
                                           [queue_nr]->userdata,
                                           pdma->dma_queue
                                           [queue_nr]->img_size, pdma->dma_queue[queue_nr]->img_id.counter);
    } else {
        dprintk_dma(MVDMA_DEBUG_TASKLET, "EOF %d without handler !\n", queue_nr);
    }
}

static void mvdma_finish_queue(struct mvdma_device *pdma, int queue_nr)
{
    if (test_and_clear_bit(queue_nr, pdma->dma_queue_deliver)) {
        mvdma_count_image(pdma, queue_nr);
        mvdma_deliver_queue(pdma, queue_nr);
    } else {
        dprintk_dma(MVDMA_DEBUG_IRQ, "queue_nr 0x%04x not in pdma->dma_queue_deliver!\n", queue_nr);
    }
}

static inline int mvdma_handle_viddma_irq(struct mvdma_device *pdma)
{
    u32 queue_nr = pdma->dma_queue_first;
    pdma->dma_desc[queue_nr]->flags_count = 0;
    /* the descriptor has been written by the CPU, so sync to device (flush cache to RAM) */
    dma_sync_single_for_device(pdma->dev, pdma->dma_handle[queue_nr], sizeof(struct mvdma_hw_desc), DMA_TO_DEVICE);
    dprintk_dma(MVDMA_DEBUG_IRQ, "dma_queue_first = 0x%04x\n", queue_nr);
    return queue_nr;
}

static inline void mvdma_next_queue(struct mvdma_device *pdma) __must_hold(&pdma->lock)
{
    dprintk_dma(MVDMA_DEBUG_IRQ, "BEFORE: %d entries / %d first\n", pdma->dma_queue_entries, pdma->dma_queue_first);

    pdma->dma_queue_entries--;
    pdma->dma_queue_first++;
    pdma->dma_queue_first %= MVDMA_QUEUE_CNT;

    dprintk_dma(MVDMA_DEBUG_IRQ, "AFTER: %d entries / %d first\n", pdma->dma_queue_entries, pdma->dma_queue_first);
}

/* callers need to have lock for the array of TmvUI_umem's */
static inline int mvdma_release_um(TmvUI_umem * um)
{
    int i;
    TmvUI_userhandle *uh = um->uh;
    struct mvdma_device *pdma = uh->pdma;

    if (!um)
        return -1;

    if (!um->page_cnt)
        return 0;

    if (!mutex_is_locked(&uh->umem_lock)) {
        pr_err("%s: ERROR! umem is NOT locked!\n", __FUNCTION__);
        BUG();
    }

    BUG_ON(um->dma_state != DMA_IDLE);

    dprintk_dma(MVDMA_DEBUG_MEM, "release %d holding %d pages\n", um->nr, um->page_cnt);

    for (i = 0; i < um->page_cnt; i++) {
        if (um->pages[i])
            put_page(um->pages[i]);
        um->pages[i] = NULL;
    }

    if (um->sl) {
        kfree(um->sl);
        um->sl = NULL;
    }
    kfree(um->pages);
    um->pages = NULL;
    um->page_cnt = 0;
    um->len = 0;
    um->uva = 0;
    um->flags = 0;

    return 0;
}

static int mvdma_um_to_sg(TmvUI_umem * um)
{
    int i;
    unsigned long bytes_left, sl_size;

    if (um->sl)
        kfree(um->sl);
    um->sl = kcalloc(um->page_cnt, sizeof(struct scatterlist), GFP_KERNEL);
    if (!um->sl) {
        pr_err("%s: ERROR! kcalloc (%d sl) failed !n", __FUNCTION__, um->page_cnt);
        return -1;
    }
    sg_init_table(um->sl, um->page_cnt);    /* GRA new! This clears the scatterlist and
                                               writes the end marker - otherwise need
                                               to use sg_mark_end somewhere */
    bytes_left = um->len;
    for (i = 0; i < um->page_cnt; i++) {
        if (i == 0) {
            sg_set_page(&um->sl[0], um->pages[0], 0, um->uva & ~PAGE_MASK);
            um->sl[0].length = PAGE_SIZE - um->sl[0].offset;
            if (um->sl[0].length > bytes_left)
                um->sl[0].length = bytes_left;
            bytes_left -= um->sl[0].length;
        } else {
            sl_size = bytes_left < PAGE_SIZE ? bytes_left : PAGE_SIZE;
            sg_set_page(&um->sl[i], um->pages[i], sl_size, 0);
            bytes_left -= sl_size;
        }
    }
    return 0;
}

static int mvdma_get_user_pages(TmvUI_userhandle * uh, unsigned long uva, size_t len)
{
    int result, nr;
    unsigned long start, end;
    TmvUI_umem *um = NULL;
    struct mvdma_device *pdma = uh->pdma;
    int found = 0;

    for (nr = 0; nr < MVDMA_QUEUE_CNT; nr++) {
        if (uh->umem[nr].len == 0) {
            uh->umem[nr].len = len;
            um = &uh->umem[nr];
            found = 1;
            break;
        }
    }
    if (!um) {
        for (nr = (uh->last_umem + 1) % MVDMA_QUEUE_CNT; nr != uh->last_umem; nr = (nr + 1) % MVDMA_QUEUE_CNT) {
            um = &uh->umem[nr];
            if (um->dma_state == DMA_IDLE) {
                mvdma_release_um(um);
                found = 1;
                break;
            }
        }
        dprintk_dma(MVDMA_DEBUG_MEM, "re-allocated umem %d\n", nr);
    }
    BUG_ON(!um);

    if (!found) {
        return -1;
    }

    um->len = len;
    um->uva = uva;
    um->offset = um->uva & ~PAGE_MASK;

    start = (um->uva & PAGE_MASK) >> PAGE_SHIFT;
    end = ((um->uva + um->len - 1) & PAGE_MASK) >> PAGE_SHIFT;

    um->page_cnt = end - start + 1;
    um->pages = kcalloc(um->page_cnt, sizeof(struct page *), GFP_KERNEL);

    dprintk_dma(MVDMA_DEBUG_MEM, "trying to map uva 0x%08lx len=0x%lx @ buffer %d\n", um->uva, um->len, nr);

    if (um->pages) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 8, 0)
        down_read(&current->mm->mmap_lock);
#else
        down_read(&current->mm->mmap_sem);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 9, 0)
        {
            int locked = 1;
            result = get_user_pages_locked(um->uva, um->page_cnt, FOLL_WRITE, um->pages, &locked);
            if (locked) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 8, 0)
                up_read(&current->mm->mmap_lock);
#else
                up_read(&current->mm->mmap_sem);
#endif
            }
        }
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 4, 0)
        result = get_user_pages(current, current->mm, um->uva, um->page_cnt, FOLL_WRITE, um->pages, NULL);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 8, 0)
        up_read(&current->mm->mmap_lock);
#else
        up_read(&current->mm->mmap_sem);
#endif
#else /* < 4.4.0 */
        result = get_user_pages(current, current->mm, um->uva, um->page_cnt, 1, 0, um->pages, NULL);
        up_read(&current->mm->mmap_sem);
#endif
        if (result == um->page_cnt) {
            dprintk_dma(MVDMA_DEBUG_MEM, "uva 0x%08lx mapped by " "%d pages \n", um->uva, um->page_cnt);
            mvdma_um_to_sg(um);
            uh->last_umem = nr;
            return 0;
        } else {
            pr_err("%s: ERROR! kmalloc pages failed (%d)\n", __FUNCTION__, result);
            mvdma_release_um(um);
            return -EFAULT;
        }
    }
    return -ENOMEM;
}

/* callers need to have lock for the array of TmvUI_umem's */
static TmvUI_umem *mvdma_find_um(TmvUI_userhandle * uh, unsigned long uva, size_t len)
{
    struct mvdma_device *pdma = uh->pdma;
    int i;

    for (i = 0; i < MVDMA_QUEUE_CNT; i++) {
        if (uh->umem[i].uva == uva) {
            if (!len || (uh->umem[i].len == len)) {
                /* SyncOnBuf (len==0) or buffer length matches
                 */
                dprintk_dma(MVDMA_DEBUG_MEM,
                            "found matching user mem %d @ 0x%lx with img_size = 0x%x\n",
                            i, uh->umem[i].uva, uh->umem[i].img_size);
                return &uh->umem[i];
            } else {
                dprintk_dma(MVDMA_DEBUG_MEM, "remap length mismatched umem %d\n", i);
                /* same pointer - different length -> remap */
                mvdma_release_um(&uh->umem[i]);
            }
        }
    }
    /* no buffer found to wait for */
    if (!len)
        return NULL;

    if (!mvdma_get_user_pages(uh, uva, len)) {
        for (i = 0; i < MVDMA_QUEUE_CNT; i++) {
            if (uh->umem[i].uva == uva) {
                return &uh->umem[i];
            }
        }
        /* buffer _must_ exist */
        BUG();
    }

    return NULL;
}

static inline void mvdma_dma_startstop(struct mvdma_device *pdma, int start, int reset) __must_hold(&pdma->lock)
{
    start = start ? 1 : 0;
    mv_dma_write(MVDMA_CTRL_WORD, start);
    if (!start && reset) {
        mv_dma_write(MVDMA_DESC_NUMBER, 0);
    }
}

int mvdma_reset(void *handle) __must_hold(&pdma->lock)
{
    int i;
    struct mvdma_device *pdma = (struct mvdma_device *)handle;
    dma_addr_t dma_handle;

    mvdma_dma_startstop(pdma, 0, 1);

    /* reset queue lists to start */
    dma_handle = pdma->dma_handle[0];
    pdma->dma_queue_entries = 0;
    pdma->dma_queue_first = 0;
    pdma->fpga_leading_by = 0;
    pdma->prepared_descriptors = 0; 

    mv_dma_write(MVDMA_DESC_RING_ADR_UPPER, UPPER_WORD(dma_handle));
    mv_dma_write(MVDMA_DESC_RING_ADR_LOWER, LOWER_WORD(dma_handle));

    /* reset all the queue parameters */
    for (i = 0; i < MVDMA_QUEUE_CNT; i++) {
        if (pdma->dma_queue[i]->valid_sgcnt) {
            dprintk_dma(MVDMA_DEBUG_MAPPING,
                        "%s: dma_unmap_sg queue:%d, valid_sgcnt=%d, sgcnt=%d\n",
                        __FUNCTION__, i, pdma->dma_queue[i]->valid_sgcnt, pdma->dma_queue[i]->sgcnt);
            /* the image buffer may have been written by the device */
            dma_unmap_sg(pdma->dev, pdma->dma_queue[i]->sg, pdma->dma_queue[i]->sgcnt, DMA_FROM_DEVICE);
        }
        pdma->dma_queue[i]->sgcnt = 0;
        pdma->dma_queue[i]->valid_sgcnt = 0;
        pdma->dma_queue[i]->cb_user = NULL;
        pdma->dma_queue[i]->userdata = NULL;
        pdma->dma_queue[i]->img_size = 0;

        mvdma_clear_pool_for_queue(pdma, i);

        pdma->dma_desc[i]->flags_count = 0;
        /* the descriptor has been written by the CPU, so sync to device (flush cache to RAM) */
        dma_sync_single_for_device(pdma->dev, pdma->dma_handle[i], sizeof(struct mvdma_hw_desc), DMA_TO_DEVICE);
    }
    bitmap_zero(pdma->dma_queue_deliver, MVDMA_QUEUE_CNT);
    clear_page_list_pool(pdma);
    return 0;
}

static int mvdma_abort(TmvUI_userhandle * uh, unsigned long uva)
{
    int i;
    unsigned long flags;
    int result = -ENOENT;
    struct mvdma_device *pdma = uh->pdma;

    /* need to get umem_lock *before* pdma->lock because
     * this is the order that happens in mvdma_queue
     */
    if (mutex_is_locked(&uh->umem_lock))
        dprintk_dma(MVDMA_DEBUG_QUEUE, "%s: umem is already locked!\n", __FUNCTION__);
    if (mutex_lock_interruptible(&uh->umem_lock)) {
        pr_err("%s: ERROR! Cannot get the umem lock!\n", __FUNCTION__);
        return -EFAULT;
    }

    if (uva) {
        dprintk_dma(MVDMA_DEBUG_QUEUE, "Looking for uva 0x%08lx\n", uva);
    }

    spin_lock_irqsave(&uh->pdma->lock, flags);
    for (i = 0; i < MVDMA_QUEUE_CNT; i++) {
        if ((0 == uva) || (uh->umem[i].uva == uva)) {
            u32 queue_nr = uh->umem[i].nr;

            if (0 != uva) {
                dprintk_dma(MVDMA_DEBUG_QUEUE, "Found uva 0x%08lx @ %d, queue %d\n", uva, i, queue_nr);
            }
            if (uh->umem[i].dma_state != DMA_IDLE) {
                dprintk_dma(MVDMA_DEBUG_MAPPING,
                            "%s: queue: %d, dma_state = 0x%x, valid_sgcnt = %d, sgcnt = %d\n",
                            __FUNCTION__, queue_nr,
                            uh->umem[i].dma_state,
                            pdma->dma_queue[queue_nr]->valid_sgcnt, pdma->dma_queue[queue_nr]->sgcnt);
                if (pdma->dma_queue[queue_nr]->valid_sgcnt) {
                    dprintk_dma(MVDMA_DEBUG_MAPPING,
                                "%s: dma_unmap_sg pdma->dma_queue[%d]->sg (%px), valid_sgcnt=%d, sgcnt=%d\n",
                                __FUNCTION__, queue_nr,
                                pdma->dma_queue[queue_nr]->sg,
                                pdma->dma_queue[queue_nr]->valid_sgcnt, pdma->dma_queue[queue_nr]->sgcnt);
                    /* the image buffer may have been written by the device */
                    dma_unmap_sg(pdma->dev,
                                 pdma->dma_queue[queue_nr]->sg, pdma->dma_queue[queue_nr]->sgcnt, DMA_FROM_DEVICE);
                    mvdma_clear_pool_for_queue(pdma, queue_nr);
                    pdma->dma_queue[queue_nr]->sgcnt = 0;
                    pdma->dma_queue[queue_nr]->valid_sgcnt = 0;
                }
                dprintk_dma(MVDMA_DEBUG_QUEUE, "Abort queue %d (set to idle).\n", i);
                uh->umem[i].dma_state = DMA_IDLE;
                mvdma_release_um(&uh->umem[i]);
                wake_up(&uh->umem[i].wq);
                result = 0;
            }
#if 0
            else {
                dprintk_dma(MVDMA_DEBUG_QUEUE, "Queue %d already idle.\n", i);
            }
#endif

            if (0 != uva) {
                /* uva was found so we are finished here */
                break;
            }
            /* else continue to abort *all* queues */
        }
    }
    spin_unlock_irqrestore(&uh->pdma->lock, flags);
    mutex_unlock(&uh->umem_lock);

    dprintk_dma(MVDMA_DEBUG_QUEUE, "Returning %d\n", result);
    return result;
}

static ssize_t mvdma_queue(struct file *filp, char *buff, size_t len, loff_t * offset)
{
    struct mvpci_fh *fh = (struct mvpci_fh *)filp->private_data;
    struct mvpci_driver *mvpci_drv = fh->mvpci_drv;
    struct mvdma_device *pdma = mvpci_drv->pdma;
    TmvUI_userhandle *uh = pdma->uh;
    TmvUI_umem *um;
    ssize_t result;
    int ret;

    if (unlikely(!uh))
        return -ENOSYS;

    /* check that we have at least one page free */
    if (unlikely(bitmap_full(pdma->page_list_pool_map, PAGES_IN_POOL))) {
        pr_err("%s: ERROR! There is no space left in the page list pool!\n", __FUNCTION__);
        return -ENOMEM;
    }

    /* Need to lock um from the time it is found to the time we're
     * done with it below. We can't allow ABORT to come in between
     * finding the um and calling mvdma_dma_to_user(um).
     */
    if (mutex_is_locked(&uh->umem_lock))
        dprintk_dma(MVDMA_DEBUG_QUEUE, "%s: umem is already locked!\n", __FUNCTION__);
    if (mutex_lock_interruptible(&uh->umem_lock)) {
        pr_err("%s: ERROR! Cannot get the umem lock!\n", __FUNCTION__);
        return -EFAULT;
    }

    um = mvdma_find_um(uh, (unsigned long)buff, len);
    if (um) {
        dprintk_dma(MVDMA_DEBUG_STD, "dma %ld bytes to user buffer %d @ 0x%px\n", len, um->nr, buff);
        result = mvdma_dma_to_user(um);
        ret = result ? result : len;
    } else {
        pr_err("%s: ERROR! Can't find free umem for 0x%px\n", __FUNCTION__, buff);
        ret = -ENOMEM;
    }

    mutex_unlock(&uh->umem_lock);

    return ret;
}

static void mvdma_irq_tasklet(unsigned long data)
{
    struct mvdma_device *pdma = (struct mvdma_device *)data;
    u32 queue_done = 0;
    u32 qbit = 0;
    u32 queue_nr = 99;
    int num_runs = 1;
    volatile u32 fpga_counter;

    spin_lock(&pdma->lock);

    fpga_counter = mv_dma_read(MVDMA_FPGA_COUNTER) >> 16;
    if (fpga_counter > MVDMA_QUEUE_CNT) {
        fpga_counter = mv_dma_read(MVDMA_FPGA_COUNTER) >> 16;
        if (fpga_counter > MVDMA_QUEUE_CNT) {
            dprintk_dma(0,
                        "%s: WARNING! SugarPlumFairies... even reading twice led to a completion timeout\n",
                        __FUNCTION__);
            fpga_counter = (pdma->dma_queue_first + 1) % MVDMA_QUEUE_CNT;
        }
    }
    if (fpga_counter < pdma->dma_queue_first) {
        pdma->fpga_leading_by = MVDMA_QUEUE_CNT + fpga_counter - pdma->dma_queue_first;
    } else {
        pdma->fpga_leading_by = fpga_counter - pdma->dma_queue_first;
    }

    num_runs = pdma->fpga_leading_by;
    if (num_runs > 1) {
        dprintk_dma(MVDMA_DEBUG_STD, "%s: WARNING! DMA/Kernel is out of sync by %d\n", __FUNCTION__, num_runs - 1);
    }

    while (num_runs-- > 0) {
        queue_nr = pdma->dma_queue_first;
        qbit = mvdma_handle_viddma_irq(pdma);
        if (pdma->dma_queue_entries > 0) {
            mvdma_next_queue(pdma);
            queue_done = qbit;

            if (queue_done < MVDMA_QUEUE_CNT) {
                set_bit(queue_done, pdma->dma_queue_deliver);
                dprintk_dma(MVDMA_DEBUG_IRQ,
                            "schedule delivery of queue 0x%04x - pending = %d\n", queue_done, pdma->dma_queue_entries);
                mvdma_finish_queue(pdma, queue_nr);
            }
        } else {
            pr_err("%s: ERROR! dma_queue_entries must be > 0!  (0x%x)\n", __FUNCTION__, pdma->dma_queue_entries);
        }
        dprintk_dma(MVDMA_DEBUG_IRQ, "%d pending\n", pdma->dma_queue_entries);

#ifdef SHOW_HUNGER
        if (!pdma->dma_queue_entries) {
            dprintk_dma(0, "HUNGER!\n");
        }
#endif
    }                           /* while */

    spin_unlock(&pdma->lock);
}

static inline int mvdma_wait_on_umem(TmvUI_umem * um, u32 timeout)
{
    TmvUI_userhandle *uh = um->uh;
    struct mvdma_device *pdma = uh->pdma;

    int result = 0;

#ifdef GET_OUT_OF_JAIL_FREE_CARD
    if (!timeout)
        timeout = 5000;
#endif // GET_OUT_OF_JAIL_FREE_CARD

    if (timeout) {
        dprintk_dma(MVDMA_DEBUG_STD, "buf %d (timeout %dms)\n", um->nr, timeout);
        wait_event_interruptible_timeout(um->wq, um->dma_state < DMA_RUNNING, msecs_to_jiffies(timeout));
    } else {
        wait_event_interruptible(um->wq, um->dma_state < DMA_RUNNING);
    }
    switch (um->dma_state) {
    case DMA_DONE:
        dprintk_dma(MVDMA_DEBUG_QUEUE, "got buf %d / img %ld\n", um->nr, uh->image_nr);
        uh->image_nr++;
        break;
    case DMA_IDLE:
        dprintk_dma(MVDMA_DEBUG_STD, "aborted buf %d\n", um->nr);
        result = -EINTR;
        break;
    default:
        dprintk_dma(MVDMA_DEBUG_STD, "buf %d timed out\n", um->nr);
        result = -ETIMEDOUT;
        break;
    }
    um->dma_state = DMA_IDLE;

    if (!result) {
        if (mutex_is_locked(&uh->umem_lock))
            dprintk_dma(MVDMA_DEBUG_QUEUE, "%s: umem is already locked!\n", __FUNCTION__);
        if (mutex_lock_interruptible(&uh->umem_lock)) {
            pr_err("%s: ERROR! Cannot get the umem lock!\n", __FUNCTION__);
            return -EFAULT;
        }
        mvdma_release_um(um);
        mutex_unlock(&uh->umem_lock);
    }

    return result;
}

static int mvdma_wait_on_buffer(TmvUI_userhandle * uh, TmvDMA_UserBuffer * ub, TmvUI_umem ** m)
{
    int result = -ENOENT;
    struct mvdma_device *pdma = uh->pdma;
    TmvUI_umem *um;

    if (mutex_is_locked(&uh->umem_lock))
        dprintk_dma(MVDMA_DEBUG_QUEUE, "%s: umem is already locked!\n", __FUNCTION__);
    if (mutex_lock_interruptible(&uh->umem_lock)) {
        pr_err("%s: ERROR! Cannot get the umem lock!\n", __FUNCTION__);
        return -EFAULT;
    }

    um = mvdma_find_um(uh, (unsigned long)(ub->buffer), 0);

    mutex_unlock(&uh->umem_lock);

    if (um) {
        *m = um;
        result = mvdma_wait_on_umem(um, ub->timeout);
        ub->img_size = um->img_size;
        ub->img_id = um->img_id;
    } else {
        pr_err("%s: ERROR! buffer '0x%pK' not found\n", __FUNCTION__, ub->buffer);
    }

    return result;
}

long mvdma_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    long result = 0;
    u32 cp_size;
    u32 io_size = _IOC_SIZE(cmd);
    u64 val;
    TmvUI_umem *um = NULL;
    struct mvpci_fh *fh = (struct mvpci_fh *)filp->private_data;

    if (fh) {
        struct mvpci_driver *mvpci_drv = fh->mvpci_drv;
        if (mvpci_drv) {
            struct mvdma_device *pdma = mvpci_drv->pdma;
            if (pdma) {
                // dprintk_dma(MVDMA_DEBUG_STD, "%s nr %d, arg 0x%lx\n", pci_name(pdma->mvpci_dev->pci), _IOC_NR(cmd), arg);

                switch (cmd) {
                case MVDMA_IOC_DMA_ENABLE:{
                        dprintk_dma(MVDMA_DEBUG_STD, "MVDMA_IOC_DMA_ENABLE: *** not implemented ***\n");
                        break;
                    }
                case MVDMA_IOC_QUEUE_BUFFER:{
                        TmvDMA_UserBuffer ub;
                        TmvDMA_UserBuffer __user *p = (TmvDMA_UserBuffer __user *) arg;

                        cp_size = CHECK_IO_SIZE(io_size, sizeof(TmvDMA_UserBuffer));
                        if ((result = copy_from_user(&ub, p, cp_size))) {
                            result = -EFAULT;
                            break;
                        }

                        dprintk_dma(MVDMA_DEBUG_STD,
                                    "MVDMA_IOC_QUEUE_BUFFER: buff=0x%px size=%ld bytes timeout=%d\n",
                                    ub.buffer, ub.size, ub.timeout);
                        result = mvdma_queue(filp, (char *)ub.buffer, ub.size, NULL);
                        break;
                    }
                case MVDMA_IOC_DEQUEUE_BUFFER:{
                        TmvDMA_UserBuffer ub;
                        long retval = 0;
                        TmvDMA_UserBuffer __user *p = (TmvDMA_UserBuffer __user *) arg;

                        cp_size = CHECK_IO_SIZE(io_size, sizeof(TmvDMA_UserBuffer));
                        if ((result = copy_from_user(&ub, p, cp_size))) {
                            result = -EFAULT;
                            break;
                        }

                        retval = mvdma_wait_on_buffer(pdma->uh, &ub, &um);
                        dprintk_dma(MVDMA_DEBUG_STD,
                                    "MVDMA_IOC_DEQUEUE_BUFFER: buff=0x%px size=%ld bytes timeout=%d, img_size=0x%x, img_id=%d\n",
                                    ub.buffer, ub.size, ub.timeout, ub.img_size, ub.img_id);

                        if (retval < 0) {
                            result = retval;
                            break;
                        }
                        put_user(ub.img_size, &p->img_size);
                        put_user(ub.img_id, &p->img_id);
                        break;
                    }
                case MVDMA_IOC_ABORT:{
                        if (__get_user(val, (uint64_t __user *) arg))
                            return -EFAULT;

                        dprintk_dma(MVDMA_DEBUG_STD, "MVDMA_IOC_ABORT: val=0x%08llx\n", val);
                        result = mvdma_abort(pdma->uh, val);    /* this takes the lock so don't hold it here! */
                        break;
                    }
                case MVDMA_IOC_RESET_DMA:{
                        /* user parameter is unused */
                        dprintk_dma(MVDMA_DEBUG_STD, "MVDMA_IOC_RESET_DMA:\n");
                        /* stop DMA running */
                        spin_lock_irq(&pdma->lock);
                        mvdma_dma_startstop(pdma, 0, 1);
                        spin_unlock_irq(&pdma->lock);

                        /* abort all queues */
                        mvdma_abort(pdma->uh, 0);   /* this takes the lock so don't hold it here! */

                        /* reset everything */
                        spin_lock_irq(&pdma->lock);
                        result = mvdma_reset(pdma);
                        spin_unlock_irq(&pdma->lock);
                        break;
                    }
                default:
                    result = -ENOENT;
                }
            }
        }
    }

    return result;
}

irqreturn_t mvdma_irq(int irq, void *context)
{
    struct mvpci_driver *mvpci_drv = (struct mvpci_driver *)context;
    struct mvdma_device *pdma = mvpci_drv->pdma;

    dprintk_dma(MVDMA_DEBUG_IRQ, "irq: %d\n", irq);

    /* just schedule the bottom-half tasklet */
    tasklet_hi_schedule(&pdma->irq_tasklet);
    return IRQ_RETVAL(1);
}

static int mvdma_init_hw(struct mvdma_device *pdma)
{
    int result = 0;
    return result;
}

static int mvdma_alloc_dma(struct mvdma_device *pdma)
{
    int qnr = 0;
    struct mvdma_hw_desc *pdesc;
    dma_addr_t dma_handle = 0;

    dprintk_dma(MVDMA_DEBUG_MAPPING, "reserving %d * 0x%lx = 0x%lx\n",
                MVDMA_QUEUE_CNT, sizeof(struct mvdma_hw_desc), MVDMA_QUEUE_CNT * sizeof(struct mvdma_hw_desc));
    pdesc = (struct mvdma_hw_desc *)dma_alloc_coherent(pdma->dev, MVDMA_QUEUE_CNT * sizeof(struct mvdma_hw_desc), &dma_handle, GFP_KERNEL);
    if (!pdesc) {
        dprintk_dma(MVDMA_DEBUG_MEM, "Descriptor ring allocation failed\n");
        return -ENOMEM;
    }
    /* calculate a set of dma handles for the descriptor ring */
    for (qnr = 0; qnr < MVDMA_QUEUE_CNT; qnr++) {
        pdma->dma_desc[qnr] = pdesc;
        pdma->dma_handle[qnr] = dma_handle;
        dprintk_dma(MVDMA_DEBUG_MEM, "dma_desc[%d] = 0x%pK, dma_handle[%d] = %pad\n", qnr, pdesc, qnr, &dma_handle);
        pdesc++;
        dma_handle += sizeof(struct mvdma_hw_desc);
    }

    /* this is where we tell the DMA where to find the queue descriptor ring */
    pdesc = pdma->dma_desc[0];
    dma_handle = pdma->dma_handle[0];
    dprintk_dma(MVDMA_DEBUG_QUEUE, "MICROSEQUENCER Start Point = %pad (kv addr = 0x%pK)\n", &dma_handle, pdesc);
    mvdma_dma_startstop(pdma, 0, 1);    /* DMA should not be running and offset set to 0 */

    mv_dma_write(MVDMA_DESC_RING_ADR_UPPER, UPPER_WORD(dma_handle));
    mv_dma_write(MVDMA_DESC_RING_ADR_LOWER, LOWER_WORD(dma_handle));

    return 0;
}

int mvdma_request(void *handle, struct scatterlist *sg, u32 sgcnt, mvdma_if_callback * cb, void *data /* = *um */ )
{
    TmvUI_umem *um = (TmvUI_umem *) data;
    struct mvdma_device *pdma = (struct mvdma_device *)handle;
    u32 valid_sgcnt, queue_nr;
    struct scatterlist *list = sg;
    struct scatterlist *sg1;
    int i, j, k;

    if (!(pdma && list && sgcnt)) {
        pr_err("%s: ERROR! pdma 0x%pK / list 0x%pK (%d entries) / userdata (um) 0x%pK\n",
               __FUNCTION__, pdma, list, sgcnt, um);
        return -EFAULT;
    }
    if (pdma->dma_queue_entries == MVDMA_QUEUE_CNT) {
        dprintk_dma(MVDMA_DEBUG_QUEUE, "request queue full !\n");
        return -EBUSY;
    }
    /* the image buffer will be writteni exclusively by the device */
    valid_sgcnt = dma_map_sg(pdma->dev, list, sgcnt, DMA_FROM_DEVICE);
    if (!valid_sgcnt) {
        pr_err("%s: ERROR! Unable to map DMA with sgcnt=0x%04x! No more IOMMU resources left?\n", __FUNCTION__, sgcnt);
        return -EBUSY;
    }

    spin_lock_irq(&pdma->lock);
    queue_nr = _get_free_queue(pdma);
    pdma->dma_queue_entries++;
    spin_unlock_irq(&pdma->lock);

    dprintk_dma(MVDMA_DEBUG_MAPPING,
                "%s: dma_map_sg pdma->dma_queue[%d]->sg (%p), valid_sgcnt=%d, sgcnt=%d\n",
                __FUNCTION__, queue_nr, list, valid_sgcnt, sgcnt);

    /* note which queue the buffer is being used at */
    um->nr = queue_nr;
    pdma->dma_queue[queue_nr]->cb_user = cb;
    pdma->dma_queue[queue_nr]->userdata = data;
    pdma->dma_queue[queue_nr]->sg = sg;
    pdma->dma_queue[queue_nr]->sgcnt = sgcnt;
    pdma->dma_queue[queue_nr]->valid_sgcnt = valid_sgcnt;
    pdma->dma_queue[queue_nr]->plrcnt = 0;
    {
        unsigned long buffer_pages_in_pool = DIV_ROUND_UP_ULL(sgcnt, ENTRIES_PER_PAGE);
        unsigned long bpip = buffer_pages_in_pool;
        int block_count = DIV_ROUND_UP_ULL(bpip, PAGES_PER_BLOCK);

        dprintk_dma(MVDMA_DEBUG_POOL, "sgcnt = 0x%x, buffer_pages_in_pool = %lu, block_count = %d\n", sgcnt,
                    buffer_pages_in_pool, block_count);

        if (block_count > 2) {
            pr_err("Error reserving page list for queue %d. Too many blocks!\n", queue_nr);
            return -ENOMEM;
        }

        /* request page lists from the pool */
        for (i = 0; (i < block_count) && (bpip > 0); i++) {
            struct page_list_req *plr = &pdma->dma_queue[queue_nr]->plr[i];
            dprintk_dma(MVDMA_DEBUG_POOL, "queue_nr = 0x%x, block_count = %d, i = %d\n", queue_nr, block_count, i);
            /* set the number of pool pages for the request */
            if (bpip > PAGES_PER_BLOCK) {
                plr->num_pages = PAGES_PER_BLOCK;
                bpip -= PAGES_PER_BLOCK;
            } else {
                plr->num_pages = (uint16_t) bpip;
            }
            /* request space in the page list pool */
            if (reserve_page_list(pdma, plr)) {
                pr_err("Error reserving page list for queue %d(%d)\n", queue_nr, i);
                return -ENOMEM;
            }
        }
        /* note the number of page lists used */
        pdma->dma_queue[queue_nr]->plrcnt = block_count;

        /* step over the buffer writing the page addresses to the page list(s) */
        {
            struct page_list_req *plr0 = &pdma->dma_queue[queue_nr]->plr[0];
            struct mvdma_page_list *pl0 = (struct mvdma_page_list *)plr0->addr;
            struct page_list_req *plr1 = &pdma->dma_queue[queue_nr]->plr[1];    /* values in plr could be undefined if block_count < 2 */
            uint64_t *p0 = (uint64_t *) plr0->addr; /* the very first entry */
            uint64_t *pn = (uint64_t *) plr0->addr + 1; /* start point skipping the first entry - will be filled with the "Trinkgeld" page later */
            uint64_t *px = pn;  /* will hold address of last entry written */
            u32 page_lists_full = 0;

            dprintk_dma(MVDMA_DEBUG_POOL, "plr0->addr = 0x%px, plr0->dma_addr = 0x%llx\n", plr0->addr, plr0->dma_addr);
            i = 1;
            for_each_sg(list, sg1, valid_sgcnt, j) {
                //dprintk_dma(MVDMA_DEBUG_POOL, "sg phys=%08llx, sg virt=0x%px sg_dma_len = 0x%x\n", sg_phys(sg1), sg_virt(sg1), sg_dma_len(sg1));
                for (k = 0; k < sg_dma_len(sg1); k += PAGE_SIZE, pn++, i++) {
                    if (!page_lists_full && (i >= ADDR_BLOCK_MAX)) {
                        dprintk_dma(MVDMA_DEBUG_POOL, "writing upper and lower parts to the next field of plr0\n");
                        dprintk_dma(MVDMA_DEBUG_POOL, "plr1->addr = 0x%px, plr1->dma_addr = 0x%llx\n", plr1->addr,
                                    plr1->dma_addr);
                        pl0->next_upper = UPPER_WORD(plr1->dma_addr);
                        pl0->next_lower = LOWER_WORD(plr1->dma_addr);
                        pn = plr1->addr;    /* continue writing here */
                        page_lists_full++;
                    }
                    /* write the dma address of one page of the image buffer to the chosen page_list */
                    *pn = sg_dma_address(sg1) + k;
                    if (*pn == 0) {
                        pr_err("%s: ERROR! dma address of page in the image buffer is 0!\n", __FUNCTION__);
                        BUG();
                    }
                    px = pn;
                    //dprintk_dma(MVDMA_DEBUG_POOL, "page_list dma_addr = 0x%llx written to 0x%px\n", *pn, pn);
                }               /* k */
            }                   /* j */

            /* special case: "Trinkgeld" page at the end of the image should be filled by the DMA first */
            /* so copy the extra entry at the very end to the skipped entry at the beginning */
            *p0 = *px;
            *px = 0;
            dprintk_dma(MVDMA_DEBUG_POOL, "page_list dma_addr[0] = *TRINKGELD* 0x%llx moved from %px to %px\n", *p0, px, p0);

            dprintk_dma(MVDMA_DEBUG_MAPPING, "mapped scatterlist with %d entr%s ->queue %d\n", sgcnt,
                        sgcnt == 1 ? "y" : "ies", queue_nr);

            /* setup one descriptor to use a page table corresponding to this buffer */
            /* N.B. the descriptor 'belongs' to the CPU and is never written by the device (see also mvdma_handle_viddma_irq) */
            pdma->dma_desc[queue_nr]->flags_count = (1 << 31) | (sgcnt << 1);
            pdma->dma_desc[queue_nr]->cmd_addr_upper = UPPER_WORD(plr0->dma_addr);
            pdma->dma_desc[queue_nr]->cmd_addr_lower = LOWER_WORD(plr0->dma_addr);

            dprintk_dma(MVDMA_DEBUG_MAPPING,
                        "Descriptor contains 0x%x entries @ dma addr upper=0x%x, lower=0x%x with %d page list(s)\n",
                        pdma->dma_desc[queue_nr]->flags_count,
                        pdma->dma_desc[queue_nr]->cmd_addr_upper,
                        pdma->dma_desc[queue_nr]->cmd_addr_lower, block_count);

            /* the queue and its descriptor have been written by the CPU, so sync for device (flush cache to RAM) */
            dma_sync_single_for_device(pdma->dev, pdma->dma_handle[queue_nr], sizeof(struct mvdma_hw_desc),
                                       DMA_TO_DEVICE);
            if (block_count == 1) {
                dma_sync_single_for_device(pdma->dev, plr0->dma_addr, sizeof(dma_addr_t) * (sgcnt - ADDR_BLOCK_MAX),
                                           DMA_TO_DEVICE);
            } else {
                dma_sync_single_for_device(pdma->dev, plr0->dma_addr, sizeof(dma_addr_t) * ADDR_BLOCK_MAX,
                                           DMA_TO_DEVICE);
                dma_sync_single_for_device(pdma->dev, plr1->dma_addr, sizeof(dma_addr_t) * (sgcnt - ADDR_BLOCK_MAX),
                                           DMA_TO_DEVICE);
            }
        }
    }

    if (mv_dma_read(MVDMA_CTRL_WORD) & 0x1) {
        /* DMA is already running - tell DMA that a new descriptor is ready to be used; the DMA will increment its count */
        mv_dma_write(MVDMA_DESC_NEW_WRITE, 1);
    } else {
        /* DMA not running yet - just count the prepared descriptor; we will tell the DMA how many descriptors to use later, when it is to be started */
        spin_lock_irq(&pdma->lock);
        pdma->prepared_descriptors++;
        spin_unlock_irq(&pdma->lock);
    }
    return 0;
}

int mvdma_probe(struct mvpci_driver *mvpci_dev)
{
    struct mvdma_device *pdma = NULL;
    TmvUI_userhandle *uh;
    int result;
    int i;

    if (!mvpci_dev) {
        pr_err("%s: ERROR! mvpci_dev is NULL!\n", __FUNCTION__);
        return -1;
    }
    pdma = (struct mvdma_device *)kzalloc(sizeof(struct mvdma_device), GFP_KERNEL);
    if (!pdma) {
        pr_err("%s: ERROR! Cannot reserve DMA structure!\n", __FUNCTION__);
        return -1;
    }

    mvpci_dev->pdma = pdma;
    pdma->mvpci_dev = mvpci_dev;
    pdma->id = mvpci_dev->nr;
    pdma->debug = debug;
    pdma->dev = &(mvpci_dev->pci->dev);
    pdma->mmio_base = mvpci_dev->mmio;

    if (!pdma->mmio_base) {
        pr_err("%s: ERROR! Couldn't remap mmio of %s\n", __FUNCTION__, pci_name(pdma->mvpci_dev->pci));
        return -1;
    } else {
        dprintk_dma(MVDMA_DEBUG_STD, "mmio: 0x%px\n", pdma->mmio_base);
    }

    spin_lock_init(&pdma->lock);

    pdma->irq_tasklet.next = NULL;
    pdma->irq_tasklet.state = 0;
    pdma->irq_tasklet.count.counter = 0;
    pdma->irq_tasklet.data = (unsigned long)pdma;
    pdma->irq_tasklet.func = mvdma_irq_tasklet;

    pdma->fpga_leading_by = 0;
    pdma->prepared_descriptors = 0; 

    result = mvdma_init_hw(pdma);

    if (result < 0) {
        pr_err("i%s: init hw failed (%d)\n", __FUNCTION__, result);
        return -1;
    }

    /* allocate dma_queues */
    for (i = 0; i < MVDMA_QUEUE_CNT; i++) {
        pdma->dma_queue[i] = (struct mvdma_queue *)kzalloc(sizeof(struct mvdma_queue), GFP_KERNEL);
        if (!pdma->dma_queue) {
            pr_err("%s: ERROR! kzalloc for mvdma_queue %d failed\n", __FUNCTION__, i);
            return -ENOMEM;
        }
    }                           /* for */
    init_page_list_pool(pdma);

    if ((result = mvdma_alloc_dma(pdma)) != 0) {
        pr_err("%s: ERROR! mvdma_alloc_dma failed (%d)\n", __FUNCTION__, result);
        return result;
    } else {
        dprintk_dma(MVDMA_DEBUG_STD, "mvdma_alloc_dma OK (%d)\n", result);
    }

    uh = kzalloc(sizeof(TmvUI_userhandle), GFP_KERNEL);
    if (!uh) {
        pr_err("%s: ERROR! kzalloc for TmvUI_userhandle failed\n", __FUNCTION__);
        return -ENOMEM;
    }

    pdma->uh = uh;

    uh->pdma = pdma;
    dprintk_dma(MVDMA_DEBUG_MEM, "umem_lock mutex_init uh @ 0x%px\n", &uh);
    dprintk_dma(MVDMA_DEBUG_MEM, "pdma @ 0x%px\n", pdma);
    mutex_init(&uh->umem_lock);
    for (i = 0; i < MVDMA_QUEUE_CNT; i++) {
        uh->umem[i].nr = i;
        uh->umem[i].uh = uh;
        init_waitqueue_head(&uh->umem[i].wq);
        mutex_init(&uh->umem[i].lock);
    }
    return 0;
}

int mvdma_remove(struct mvpci_driver *mvpci_dev)
{
    if (!mvpci_dev) {
        pr_err("%s: ERROR! mvpci_dev is NULL!\n", __FUNCTION__);
        return -1;
    }

    {
        struct mvdma_device *pdma = mvpci_dev->pdma;
        int qnr;
        TmvUI_userhandle *uh = NULL;

        if (!pdma) {
            pr_err("%s: ERROR! pdma is NULL!\n", __FUNCTION__);
            return -1;
        }
        dprintk_dma(MVDMA_DEBUG_STD, "+ %s (%d)\n", __FUNCTION__, pdma->id);

        uh = pdma->uh;
        BUG_ON(!uh);

        /* stop DMA running */
        spin_lock_irq(&pdma->lock);
        mvdma_dma_startstop(pdma, 0, 1);
        spin_unlock_irq(&pdma->lock);

        /* abort all queues */
        mvdma_abort(pdma->uh, 0);   /* this takes the lock so don't hold it here! */

        flush_scheduled_work(); /* ?? */

        for (qnr = 0; qnr < MVDMA_QUEUE_CNT; qnr++) {
            mvdma_wait_on_umem(&uh->umem[qnr], 20);
        }

        /* reset everything */
        spin_lock_irq(&pdma->lock);
        mvdma_reset(pdma);
        spin_unlock_irq(&pdma->lock);

        kfree(uh);

        if (pdma->dma_desc[0]) {
            dprintk_dma(MVDMA_DEBUG_MAPPING, "freeing the descriptor ring (dma_desc[0] = 0x%pK)\n", pdma->dma_desc[0]);
            dprintk_dma(MVDMA_DEBUG_MAPPING, "freeing the descriptor ring (dma_handle[0] = 0x%08llx)\n", pdma->dma_handle[0]);
            dma_free_coherent(pdma->dev,
                              MVDMA_QUEUE_CNT * sizeof(struct mvdma_hw_desc), pdma->dma_desc[0], pdma->dma_handle[0]);
            pdma->dma_desc[0] = NULL;
        }

        free_page_list_pool(pdma);

        dprintk_dma(MVDMA_DEBUG_MEM, "freeing dma_queue\n");
        for (qnr = 0; qnr < MVDMA_QUEUE_CNT; qnr++) {
            if (pdma->dma_queue[qnr]) {
                kfree(pdma->dma_queue[qnr]);
            }
        }

        kfree(pdma);
        mvpci_dev->pdma = NULL;

        dprintk_dma(MVDMA_DEBUG_STD, "- %s\n", __FUNCTION__);
    }
    return 0;
}
