/*
 *  mvPCIe Driver - PCIe implementation
 *  Copyright (c) 2019-2022 by MATRIX VISION GmbH <info@matrix-vision.de>
 *  Author: grayh
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/errno.h>
#include <linux/uaccess.h>
#include <linux/version.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/sysfs.h>
#include <linux/aer.h>

#include "mvdma.h"

static u32 number_of_devices = 0;
static dev_t mvpci_devnum;
static struct kobject *mvpci_kobj;

u32 debug = 0;
u32 isr_debug = 0;
static u32 num_cameras = 64;

module_param(debug, int, 0644);
MODULE_PARM_DESC(debug, "configure general debug level");
module_param(isr_debug, int, 0644);
MODULE_PARM_DESC(isr_debug, "configure isr debug level");
module_param(num_cameras, int, 0644);
MODULE_PARM_DESC(num_cameras, "max. number of cameras connected (default=64)");

MODULE_DESCRIPTION(MVPCI_DRIVER_NAME " driver");
MODULE_AUTHOR("MATRIX VISION GmbH <info@matrix-vision.de>");
MODULE_LICENSE("GPL");

//#define USE_NUMA
#define USE_COHERENT_DMA
#define TRY_MSI
//#define TRY_MSIX /* not yet implemented in FPGA */
#define USE_PCI_ERROR_HANDLERS  /* not yet fully implemented */
#define COWARDLY_REFUSAL

#define _kB (1024)
#define _MB (_kB * _kB)

static struct mvpci_driver **mvpci_devices;
static struct mvpci_sysfs **mvpci_sysfs_entries;

/* ===================================== Interrupts ==========================
 */

#define INT_MSEQ   0
#define INT_IO     1
#define INT_STAT   2
#define INT_SC     3

typedef irqreturn_t(*irq_uio_handler_t) (int, struct uio_info *);

#define DEFINE_UIO_MSI_ISR(n)                               \
static irqreturn_t mvpci_isr_uio ## n (int irq, struct uio_info *info) \
{                                                           \
    struct mvpci_driver *mvpci_drv = (struct mvpci_driver *)info->priv; \
    struct mvpci_fh *fh = &mvpci_drv->fh;                   \
                                                            \
    /* device not open, ignore interrupt but mark it as processed,  \
     * as it definitely comes from our hardware but cannot be correctly used (yet)  \
     */                                                     \
    if ((number_of_devices == 0) || !mvpci_drv->initialized || !fh || (fh->opened == 0)) { \
        iprintk(0, "Spurious INTERRUPT UIO ISR irq=%d!\n", irq); \
        return IRQ_RETVAL(1);                               \
    }                                                       \
    iprintk(0x200, "INTERRUPT UIO ISR irq=%d, name=%s\n", irq, info->name);      \
    return IRQ_RETVAL(1);                                   \
}

DEFINE_UIO_MSI_ISR(INT_IO)
    DEFINE_UIO_MSI_ISR(INT_STAT)
    DEFINE_UIO_MSI_ISR(INT_SC)
#define MSI_UIO_ISR(n) mvpci_isr_uio ## n
/* interrupt handler used for MSeq with MSI and MSI-X */
static irqreturn_t mvpci_isr_mseq(int irq, void *context)
{
    struct mvpci_driver *mvpci_drv = (struct mvpci_driver *)context;
    struct mvpci_fh *fh = &mvpci_drv->fh;
    int handled = 0;

    /* device not open, ignore interrupt but mark it as processed,
     * as it definitely comes from our hardware but cannot be correctly used (yet)
     */
    if ((number_of_devices == 0) || !mvpci_drv->initialized || !fh || (fh->opened == 0)) {
        iprintk(0, "Spurious INTERRUPT MSEQ ISR irq=%d!\n", irq);
        return IRQ_RETVAL(1);
    }

    iprintk(0x200, "MSeq INTERRUPT ISR irq=%d\n", irq);
    if (mvdma_irq(irq, context) == IRQ_HANDLED) {
        handled = 1;
    }
    return IRQ_RETVAL(handled);
}

/* interrupt handler used when MSI vectors not possible i.e. one global handler
 * for everything */
static irqreturn_t mvpci_isr_single(int irq, void *context)
{
    int handled = 0;
    struct mvpci_driver *mvpci_drv = (struct mvpci_driver *)context;
    struct mvpci_fh *fh = &mvpci_drv->fh;
    u32 irq_source;

    /* device not open, ignore interrupt and mark it as NOT processed,
     * because this might be a genuine shared interrupt from different hardware!
     */
    if ((number_of_devices == 0) || !mvpci_drv->initialized || !fh || (fh->opened == 0)) {
        iprintk(0, "Spurious INTERRUPT irq=%d!\n", irq);
        return IRQ_RETVAL(0);
    }

    /* read the source of the interrupt */
    irq_source = mvread(MVPCI_REMEMBERED_IRQS) & 0xF;

    /* if any bit is set it must be one of our interrupts */
    handled = (irq_source > 0);

    if (handled) {
        iprintk(0x200, "irq_source=0x%x\n", irq_source);
        if (irq_source & IRQ_SOURCE_MSEQ) {
            /* microsequencer interrupt - call the MSeq handler directly */
            mvpci_isr_mseq(irq, context);
        }

        if (irq_source & IRQ_SOURCE_1) {
            /* I/O interrupt - trigger UIO */
            if (mvpci_drv->uio_info[INT_IO].name) {
                uio_event_notify(&mvpci_drv->uio_info[INT_IO]);
            }
        }

        if (irq_source & IRQ_SOURCE_2) {
            /* statistic interrupt - trigger UIO */
            if (mvpci_drv->uio_info[INT_STAT].name) {
                uio_event_notify(&mvpci_drv->uio_info[INT_STAT]);
            }
        }

        if (irq_source & IRQ_SOURCE_3) {
            /* statistic interrupt - trigger UIO */
            if (mvpci_drv->uio_info[INT_SC].name) {
                uio_event_notify(&mvpci_drv->uio_info[INT_SC]);
            }
        }
    }

    return IRQ_RETVAL(handled);
}

/* ===================================== Helpers =============================
 */
static int set_allocate_name(const char **pname, const char *fmt, ...)
{
    const char *s;
    va_list vargs;

    va_start(vargs, fmt);
    s = kvasprintf_const(GFP_KERNEL, fmt, vargs);
    va_end(vargs);
    if (!s)
        return -ENOMEM;
    *pname = s;
    return 0;
}

/* ===================================== PCI INIT ========================== */
static void mvinterrupts_enable_all(struct mvpci_driver *mvpci_drv, unsigned enable)
{
    int i, irq;

    if (mvpci_drv->flags & MVPCI_FLAG_MSI) {
        for (i = 0; i < MVPCI_MSI_MAX_VECTORS; i++) {
            switch (i) {
            case INT_MSEQ:
            case INT_IO:
            case INT_STAT:
            case INT_SC:
                irq = mvpci_drv->msi_x_entry[i].vector;
                if (irq && (irq != UIO_IRQ_CUSTOM)) {
                    if (enable) {
                        dprintk(1, "Enable IRQ %d\n", irq);
                        enable_irq(irq);
                    } else {
                        dprintk(1, "Disable IRQ %d\n", irq);
                        disable_irq(irq);
                    }
                }
            }                   /* switch */
        }                       /* for */
    }
}

static void mvpci_free_pci(struct mvpci_driver *mvpci_drv)
{
    int i;

    if (!mvpci_drv || !mvpci_drv->initialized) {
        return;
    }
    /* remove DMA engine */
    mvdma_remove(mvpci_drv);

    /* unregister MSeq interrupt */
    if (mvpci_drv->msi_x_entry[INT_MSEQ].vector) {
        dprintk(0, "device mvpci-%d: freeing MSI vector %d\n", mvpci_drv->nr, mvpci_drv->msi_x_entry[INT_MSEQ].vector);
        free_irq(mvpci_drv->msi_x_entry[INT_MSEQ].vector, mvpci_drv);
        mvpci_drv->msi_x_entry[INT_MSEQ].vector = 0;
    }

    /* unregister UIO devices, removing interrupts */
    for (i = INT_IO; i <= INT_SC; i++) {
        if (mvpci_drv->uio_info[i].name) {
            if (mvpci_drv->msi_x_entry[i].vector) {
                dprintk(0, "device mvpci-%d: freeing UIO vector %d\n", mvpci_drv->nr, mvpci_drv->msi_x_entry[i].vector);
                uio_unregister_device(&mvpci_drv->uio_info[i]);
                // ... and free the name which we allocated dynamically
                kfree_const(mvpci_drv->uio_info[i].name);
                mvpci_drv->uio_info[i].name = "";
                mvpci_drv->msi_x_entry[i].vector = 0;
            }
        }
    }

    /* unregister all IRQs from device */
    mvpci_drv->flags &= ~(MVPCI_FLAG_IRQ | MVPCI_FLAG_MSI);
    pci_free_irq_vectors(mvpci_drv->pci);

#ifdef USE_IOW_SEQUENCER
    if (mvpci_drv->iow_sequencer) {
        dprintk(1, "device mvpci-%d: free io_writer sequencer memory\n", mvpci_drv->nr);
        dma_free_coherent(&mvpci_drv->pci->dev,
                          MVPCI_IOW_SEQUENCER_SIZE, mvpci_drv->iow_sequencer, mvpci_drv->dma_iow_sequencer);
        mvpci_drv->iow_sequencer = NULL;
        mvpci_drv->dma_iow_sequencer = 0;
    }
#endif // USE_IOW_SEQUENCER

    if (mvpci_drv->mmio) {
        dprintk(1, "device mvpci-%d: pci_iounmap MMIO\n", mvpci_drv->nr);
        pci_iounmap(mvpci_drv->pci, mvpci_drv->mmio);
        mvpci_drv->mmio = NULL;
    }

    if (mvpci_drv->flags & MVPCI_FLAG_RES0) {
        dprintk(1, "device mvpci-%d: pci_release_regions\n", mvpci_drv->nr);
        pci_release_regions(mvpci_drv->pci);
        mvpci_drv->flags &= ~MVPCI_FLAG_RES0;
    }

    pci_set_drvdata(mvpci_drv->pci, NULL);

    dprintk(0, "device mvpci-%d: disable PCI device\n", mvpci_drv->nr);
    pci_disable_device(mvpci_drv->pci);
    mvpci_drv->initialized = 0;
}

static int mvpci_open(struct inode *inode, struct file *file)
{
    unsigned int minor = MINOR(inode->i_rdev);

    dprintk(2, "+ %s, minor = %d\n", __FUNCTION__, minor);

    if (minor < num_cameras) {
        unsigned n = minor;     /* device number */
        struct mvpci_driver *mvpci_drv = mvpci_devices[n];
        struct mvpci_fh *fh = &mvpci_drv->fh;
        struct mvdma_device *pdma = mvpci_drv->pdma;

        if (!mvpci_drv->initialized) {
            pr_err("%s: ERROR: device %d is NOT initialised!\n", __FUNCTION__, mvpci_drv->nr);
            return -EIO;
        }

        file->private_data = fh;
        fh->opened++;
        dprintk(2, "  %s:, device number = %d, opened = %d\n", __FUNCTION__, n, fh->opened);

        if (fh->opened == 1) {
            /* DMA should not be running and offset set to 0 if this is the first time opened */
            dprintk(2, "%s: reset queue and descriptor counters to 0\n", __FUNCTION__);
            spin_lock_irq(&pdma->lock);
            mvdma_reset(pdma);
            spin_unlock_irq(&pdma->lock);

            /* set up the max. burst size the first time the device is opened (just to be sure) */
            if (mvpci_drv->maxreadreq >= 128) {
                mvwrite(mvpci_drv->maxreadreq >> 3, MVDMA_CTRL_MAX_BURST_SIZE);
            } else {
                /* MaxReadReq somehow unknown, set burst to minimum (for 128 bytes) */
                mvwrite(16, MVDMA_CTRL_MAX_BURST_SIZE);
            }
            /* enable interrupts on the CPU */
            mvinterrupts_enable_all(mvpci_drv, 1);
            /* enable Interrupts from the hardware */
            mvwrite(0x04, MVPCI_IO_CONTROL);
        }
    }

    dprintk(2, "- %s\n", __FUNCTION__);
    return 0;
}

static int mvpci_release(struct inode *inode, struct file *file)
{
    dprintk(2, "+ %s\n", __FUNCTION__);

    if (file->private_data) {
        struct mvpci_fh *fh = (struct mvpci_fh *)file->private_data;

        if (fh) {
            struct mvpci_driver *mvpci_drv = fh->mvpci_drv;

            if (fh->opened)
                fh->opened--;

            dprintk(2, "  %s:, device number = %d, opened = %d\n", __FUNCTION__, mvpci_drv->nr, fh->opened);

            if (mvpci_drv) {
                struct mvdma_device *pdma = mvpci_drv->pdma;
                if (pdma) {
                    if (fh->opened == 0) {
                        /* DMA should not be running and offset set to 0 if this is the last time closed */
                        dprintk(2, "%s: reset queue and descriptor counters to 0\n", __FUNCTION__);
                        spin_lock_irq(&pdma->lock);
                        mvdma_reset(pdma);
                        spin_unlock_irq(&pdma->lock);
                        /* disable Interrupts from the hardware */
                        mvwrite(0x00, MVPCI_IO_CONTROL);
                        /* disable interrupts on the CPU */
                        mvinterrupts_enable_all(mvpci_drv, 0);
                    }
                }
            }
        }
    }
    file->private_data = NULL;

    dprintk(2, "- %s\n", __FUNCTION__);
    return 0;
}

#ifdef USE_IOW_SEQUENCER
static ssize_t mvpci_write_seq(struct file *file, const char __user * buf,
                               size_t count, loff_t * ppos) __must_hold(&mvpci_drv->dev_lock)
{
    int result = 0;
    u64 addr = (u64) (*ppos) - MVPCI_IOW_SEQUENCER_OFFSET;
    struct mvpci_fh *fh = (struct mvpci_fh *)file->private_data;
    struct mvpci_driver *mvpci_drv = fh->mvpci_drv;

    if ((addr + count) <= MVPCI_IOW_SEQUENCER_SIZE) {
        u8 *p = (u8 *) mvpci_drv->iow_sequencer + addr;
        if (copy_from_user(p, buf, count)) {
            result = -EFAULT;
            goto out;
        }
    } else {
        pr_err
            ("%s: ERROR: write length (0x%zx) invalid. Must be <= 0x%llx.\n",
             __FUNCTION__, count, MVPCI_IOW_SEQUENCER_SIZE);
        result = -EINVAL;
        goto out;
    }
    *ppos += count;
    result = count;
#if 0
    dprintk(2, "%s: seq write to addr=0x%llx, count=0x%zx, result=0x%x\n", __FUNCTION__, addr, count, result);
#endif
out:
    return result;
}
#endif // USE_IOW_SEQUENCER

static ssize_t mvpci_write(struct file *file, const char __user * buf, size_t count, loff_t * ppos)
{
#define MAX_WRITE_LEN 256       /* some arbitrary upper limit */
    const int nibble = (int)sizeof(u32);
    int result = 0;
    u64 addr = (u64) (*ppos);
    struct mvpci_fh *fh = (struct mvpci_fh *)file->private_data;
    struct mvpci_driver *mvpci_drv = fh->mvpci_drv;
    int i;
    int written = 0;
    u32 val32;
    char __user *p = (char __user *)buf;

    if (mutex_lock_interruptible(&mvpci_drv->dev_lock)) {
        return -ERESTARTSYS;
    }

    if ((addr + count) > pci_resource_len(mvpci_drv->pci, 0)) {
        pr_err
            ("%s: ERROR: Offset 0x%llx is outside range of BAR 0 (0x%llx).\n",
             __FUNCTION__, addr, pci_resource_len(mvpci_drv->pci, 0));
        result = -EINVAL;
        goto out;
    }

#ifdef USE_IOW_SEQUENCER
    if (addr >= MVPCI_IOW_SEQUENCER_OFFSET) {
        /* writing to io_writer sequencer memory */
        result = mvpci_write_seq(file, buf, count, ppos);
        goto out;
    }
#endif // USE_IOW_SEQUENCER

    /* we will process a maximum of MAX_WRITE_LEN bytes this time around,
     * but the system is allowed to call us again.
     */
    if (count > MAX_WRITE_LEN) {
        count = MAX_WRITE_LEN;
    }
    // count must be a multiple of 4 - ignore any extra bytes requested!
    count = count & 0xfffc;

    // break down into 32-bit writes
    for (i = count; (i > 0) && !result; i -= nibble, addr += nibble, p += nibble) {
        if (copy_from_user(&val32, p, nibble)) {
            result = -EFAULT;
            goto out;
        }
        if ((MVDMA_CTRL_WORD == addr) && (val32 != 0))
        {
            struct mvdma_device *pdma = mvpci_drv->pdma;
            int j;

            spin_lock_irq(&pdma->lock);
            for (j = 0; j < pdma->prepared_descriptors; j++) {
                mv_dma_write(MVDMA_DESC_NEW_WRITE, 1);
            }
            pdma->prepared_descriptors = 0;
            spin_unlock_irq(&pdma->lock);
        }
        mvwrite(val32, addr);
        written += nibble;
    }                           // for
    result = written;
out:
#if 0
    dprintk(2, "%s: wrote to addr=0x%llx, written/result=0x%x, result=0x%x\n", __FUNCTION__, addr, result);
#endif
    *ppos += written;
    mutex_unlock(&mvpci_drv->dev_lock);
    return result;
}

#ifdef USE_IOW_SEQUENCER
static ssize_t mvpci_read_seq(struct file *file, char __user * buf,
                              size_t count, loff_t * ppos) __must_hold(&mvpci_drv->dev_lock)
{
    int result = 0;
    u64 addr = (u64) (*ppos) - MVPCI_IOW_SEQUENCER_OFFSET;
    struct mvpci_fh *fh = (struct mvpci_fh *)file->private_data;
    struct mvpci_driver *mvpci_drv = fh->mvpci_drv;

    if ((addr + count) <= MVPCI_IOW_SEQUENCER_SIZE) {
        u8 *p = (u8 *) mvpci_drv->iow_sequencer + addr;
        if (copy_to_user(buf, p, count)) {
            result = -EFAULT;
            goto out;
        }
    } else {
        pr_err
            ("%s: ERROR: read length (0x%zx) invalid. Must be <= 0x%llx.\n",
             __FUNCTION__, count, MVPCI_IOW_SEQUENCER_SIZE);
        result = -EINVAL;
        goto out;
    }
    *ppos += count;
    result = count;
#if 0
    dprintk(2, "%s: seq read from addr=0x%llx, count=0x%zx, result=0x%x\n", __FUNCTION__, addr, count, result);
#endif
out:
    return result;
}
#endif // USE_IOW_SEQUENCER

static ssize_t mvpci_read(struct file *file, char __user * buf, size_t count, loff_t * ppos)
{
#define MAX_READ_LEN (4 * _kB)  /* some arbitrary upper limit */
    int result = 0;
    u64 addr = (u64) (*ppos);
    struct mvpci_fh *fh = (struct mvpci_fh *)file->private_data;
    struct mvpci_driver *mvpci_drv = fh->mvpci_drv;
    int i;
    int bytes_read = 0;
    u64 val64;
    u32 val32;
    char __user *p = buf;

    if (mutex_lock_interruptible(&mvpci_drv->dev_lock)) {
        return -ERESTARTSYS;
    }

    if ((addr + count) > pci_resource_len(mvpci_drv->pci, 0)) {
        pr_err
            ("%s: ERROR: Offset 0x%llx is outside range of BAR 0 (0x%llx).\n",
             __FUNCTION__, addr, pci_resource_len(mvpci_drv->pci, 0));
        result = -EINVAL;
        goto out;
    }

#ifdef USE_IOW_SEQUENCER
    if (addr >= MVPCI_IOW_SEQUENCER_OFFSET) {
        /* reading from io_writer sequencer memory */
        result = mvpci_read_seq(file, buf, count, ppos);
        goto out;
    }
#endif // USE_IOW_SEQUENCER

    /* we will process a maximum of MAX_READ_LEN bytes this time around,
     * but the system is allowed to call us again.
     */
    if (count > MAX_READ_LEN) {
        count = MAX_READ_LEN;
    }
    // count must be a multiple of 4 - ignore any extra bytes requested!
    count = count & 0xfffc;

    // break down into 64-bit reads with a final 32 bit read if necessary
    for (i = count; (i > 0) && !result; i -= sizeof(u64)) {
        unsigned nibble = i >= sizeof(u64) ? sizeof(u64) : sizeof(u32);
        ktime_t remaining;

        /* read timeout 8 ms */
        hrtimer_start(&mvpci_drv->read_hrtimer, ns_to_ktime(8ULL * NSEC_PER_MSEC), HRTIMER_MODE_REL);

        switch (nibble) {
        case sizeof(u64):
            val64 = mvread64(addr);
            remaining = hrtimer_get_remaining(&mvpci_drv->read_hrtimer);
            hrtimer_cancel(&mvpci_drv->read_hrtimer);
            /* If an error occurs it is defined by a timeout -> long execution time and a return value of 0xffffffff */
            if (unlikely(ktime_to_ns(remaining) <= 0)) {
                //dprintk(0, "timer has timed out! \n");
                if (val64 == 0xffffffffUL) {
                    mvpci_drv->read_timeout_count++;
                    dprintk(0, "%d: pcie register read timeout (0x%08llx), size=%d, timeout count=%llu\n",
                            mvpci_drv->nr, addr, nibble, mvpci_drv->read_timeout_count);
                    /* read 64 again */
                    val64 = mvread64(addr);
                }
            }
            if (copy_to_user(p, &val64, sizeof(u64))) {
                result = -EFAULT;
                goto out;
            }
            break;
        case sizeof(u32):
            val32 = mvread(addr);
            remaining = hrtimer_get_remaining(&mvpci_drv->read_hrtimer);
            hrtimer_cancel(&mvpci_drv->read_hrtimer);
            /* If an error occurs it is defined by a timeout -> long execution time and a return value of 0xffffffff */
            if (unlikely(ktime_to_ns(remaining) <= 0)) {
                //dprintk(0, "timer has timed out! \n");
                if (val32 == 0xffffffff) {
                    mvpci_drv->read_timeout_count++;
                    dprintk(0, "%d: pcie register read timeout (0x%08llx), size=%d, timeout count=%llu\n",
                            mvpci_drv->nr, addr, nibble, mvpci_drv->read_timeout_count);
                    /* read 32 again */
                    val32 = mvread(addr);
                }
            }
            if (copy_to_user(p, &val32, sizeof(u32))) {
                result = -EFAULT;
                goto out;
            }
            break;
        }                       // switch
        addr += nibble;
        p += nibble;
        bytes_read += nibble;
    }                           // for
    result = bytes_read;
out:
#if 0
    dprintk(2, "%s: read from addr=0x%llx, read/result=0x%x, result=0x%x\n", __FUNCTION__, addr, result);
#endif
    *ppos += bytes_read;
    mutex_unlock(&mvpci_drv->dev_lock);
    return result;
}

static int read_maxpayload(struct pci_dev *dev)
{
    int pos;
    u16 dctl, dcap, dsz, dmax;

    pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
    if (!pos)
        return -1;

    /* Read Device MaxPayload capability and current setting */
    pci_read_config_word(dev, pos + PCI_EXP_DEVCTL, &dctl);
    pci_read_config_word(dev, pos + PCI_EXP_DEVCAP, &dcap);
    dsz = (dctl & PCI_EXP_DEVCTL_PAYLOAD) >> 0x05;
    dmax = (dcap & PCI_EXP_DEVCAP_PAYLOAD);
    // only 3 values are possible (0:128, 1:256, 2:512 bytes)
    if ((dsz < 0x03) && (dmax < 0x03)) {
        dprintk(0, "MaxPayload: used = %d Bytes, capability = %d Bytes\n", 0x80 * (0x1 << dsz), 0x80 * (0x1 << dmax));
        return 0x80 * (0x1 << dsz);
    }
    dprintk(0, "MaxPayload: *UNKNOWN*!\n");
    return -1;
}

static int read_maxreadreq(struct pci_dev *dev)
{
    int pos;
    u16 dctl, dsz;

    pos = pci_find_capability(dev, PCI_CAP_ID_EXP);
    if (!pos)
        return -1;

    /* Read Device MaxReadReq current setting */
    pci_read_config_word(dev, pos + PCI_EXP_DEVCTL, &dctl);
    dsz = (dctl & PCI_EXP_DEVCTL_READRQ) >> 0x0C;
    // only 6 values are possible (0:128, 1:256, 2:512, 3:1024, 4:2048, 5:4096 bytes) but we only allow <= 1024 to be used
    if (dsz > 0x03) {
        dsz = 0x3;
    }
    dprintk(0, "MaxReadReq: using %d Bytes\n", 0x80 * (0x1 << dsz));
    return 0x80 * (0x1 << dsz);
}

#ifndef PCI_EXP_LNKSTA_CLS_16_0GB
#define PCI_EXP_LNKSTA_CLS_16_0GB 0x0004    /* Current Link Speed 16.0GT/s */
#endif

#ifndef PCI_EXP_LNKSTA_CLS_32_0GB
#define PCI_EXP_LNKSTA_CLS_32_0GB 0x0005    /* Current Link Speed 32.0GT/s */
#endif

static u32 get_pcie_link_speed(struct mvpci_driver *mvpci_drv)
{
    u16 linkstat;
    u32 speed = 2500;           /* assume Gen 1 */
    int ret;

    ret = pcie_capability_read_word(mvpci_drv->pci, PCI_EXP_LNKSTA, &linkstat);
    if (ret) {
        dprintk(0, "%s: WARNING! Unable to read the PCIe configuration! Assume PCIe Gen1.\n", __FUNCTION__);
        return speed;
    }

    switch (linkstat & PCI_EXP_LNKSTA_CLS) {
    case PCI_EXP_LNKSTA_CLS_5_0GB:
        speed = 5000;           /* Gen 2, 5GHz */
        break;
    case PCI_EXP_LNKSTA_CLS_8_0GB:
        speed = 8000;           /* Gen 3, 8GHz */
        break;
    case PCI_EXP_LNKSTA_CLS_16_0GB:
        speed = 16000;          /* Gen 4, 16GHz */
        break;
    case PCI_EXP_LNKSTA_CLS_32_0GB:
        speed = 32000;          /* Gen ?, 32GHz */
        break;
    case PCI_EXP_LNKSTA_CLS_2_5GB:
    default:
        speed = 2500;           /* Gen 1, 2.5GHz */
        break;
    }

    return speed;
}

static u32 get_pcie_link_width(struct mvpci_driver *mvpci_drv)
{
    u16 linkstat;
    int ret;

    ret = pcie_capability_read_word(mvpci_drv->pci, PCI_EXP_LNKSTA, &linkstat);
    if (ret) {
        dprintk(0, "%s: WARNING! Unable to read the PCIe configuration! Assume x1.\n", __FUNCTION__);
        return 1;               /* assume Gen1 */
    }

    return (linkstat & PCI_EXP_LNKSTA_NLW) >> PCI_EXP_LNKSTA_NLW_SHIFT;
}

#define CONVERT_PCIE_SPEED_TO_MBS(speed) \
      ((speed) == 32000 ? 32000*128/130 : \
      (speed) == 16000 ? 16000*128/130 : \
      (speed) == 8000  ?  8000*128/130 : \
      (speed) == 5000  ?  5000*8/10 : \
      (speed) == 2500  ?  2500*8/10 : \
      0)

static u32 get_pcie_link_bandwidth(struct mvpci_driver *mvpci_drv)
{
    u32 link_width;
    u32 link_speed;
    u32 link_bandwidth;

    link_speed = get_pcie_link_speed(mvpci_drv);
    link_width = get_pcie_link_width(mvpci_drv);
    link_bandwidth = link_width * CONVERT_PCIE_SPEED_TO_MBS(link_speed);

    return link_bandwidth;
}

static long mvpci_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    long result = 0;
    int err = 0;
    u32 cp_size;
    u32 io_size = _IOC_SIZE(cmd);
    u32 arg32;
    struct mvpci_fh *fh = (struct mvpci_fh *)filp->private_data;
    struct mvpci_driver *mvpci_drv = fh->mvpci_drv;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
    err = !access_ok((void __user *)arg, _IOC_SIZE(cmd));
#else
    if (_IOC_DIR(cmd) & _IOC_READ)
        err = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
    else if (_IOC_DIR(cmd) & _IOC_WRITE)
        err = !access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
#endif
    if (err)
        return -EFAULT;

    if (fh) {
        switch (cmd) {
        case MVPCI_IOC_WRITE_BAR:{
                TmvPCI_Register32 reg;
                cp_size = CHECK_IO_SIZE(io_size, sizeof(TmvPCI_Register32));
                if ((result = copy_from_user(&reg, (TmvPCI_Register32 __user *)
                                             arg, cp_size))) {
                    result = -EFAULT;
                    break;
                }

                dprintk(MVPCI_FLAG_IORW,
                        "device mvpci-%d: MVPCI_IOC_WRITE_BAR: offset=0x%04x value=0x%04x size=%ld bytes\n",
                        mvpci_drv->nr, reg.offset, reg.value, reg.size);
                mvwrite(reg.value, reg.offset);
                break;
            }
        case MVPCI_IOC_READ_BAR:{
                TmvPCI_Register32 reg;
                long retval = 0;
                cp_size = CHECK_IO_SIZE(io_size, sizeof(TmvPCI_Register32));
                if ((result = copy_from_user(&reg, (TmvPCI_Register32 __user *)
                                             arg, cp_size))) {
                    result = -EFAULT;
                    break;
                }

                retval = mvread(reg.offset);

                if (retval < 0) {
                    result = retval;
                    break;
                }
                reg.value = retval;
                dprintk(MVPCI_FLAG_IORW,
                        "device mvpci-%d: MVPCI_IOC_READ_BAR: offset=0x%04x value=0x%04x size=%ld bytes\n",
                        mvpci_drv->nr, reg.offset, reg.value, reg.size);

                result = copy_to_user((TmvPCI_Register32 __user *)
                                      arg, &reg, sizeof(reg));
                break;
            }
        case MVPCI_IOC_READ_MAXPAYLOAD:{
                dprintk(1, "device mvpci-%d: MVPCI_READ_MAXPAYLOAD: %d\n", mvpci_drv->nr, mvpci_drv->maxpayloadsize);
                result = __put_user(mvpci_drv->maxpayloadsize, (uint32_t __user *) arg);
                break;
            }
        case MVPCI_IOC_READ_MAXREADREQ:{
                dprintk(1, "device mvpci-%d: MVPCI_READ_MAXREADREQ: %d\n", mvpci_drv->nr, mvpci_drv->maxreadreq);
                result = __put_user(mvpci_drv->maxreadreq, (uint32_t __user *) arg);
                break;
            }
        case MVPCI_IOC_WRITE_USER_INFO:{
                TmvPCI_user_info user_info;
                if (copy_from_user(&user_info, (uint64_t __user *) arg, sizeof(TmvPCI_user_info)))
                    return -EFAULT;
                dprintk(1, "device mvpci-%d: MVPCI_IOC_WRITE_USER_INFO: count=0x%x, owner=0x%x\n",
                        mvpci_drv->nr, user_info.user_count, user_info.streaming_owner);
                mvpci_drv->streaming_owner = user_info.streaming_owner;
                mvpci_drv->user_count = user_info.user_count;
                break;
            }
        case MVPCI_IOC_READ_USER_INFO:{
                TmvPCI_user_info user_info;
                user_info.streaming_owner = mvpci_drv->streaming_owner;
                user_info.user_count = mvpci_drv->user_count;
                dprintk(1, "device mvpci-%d: MVPCI_IOC_READ_USER_INFO: count=0x%x, owner=0x%x\n",
                        mvpci_drv->nr, user_info.user_count, user_info.streaming_owner);
                result = copy_to_user((u64 __user *) arg, &user_info, sizeof(TmvPCI_user_info));
                break;
            }
        case MVPCI_IOC_WRITE_MAXPAYLOAD:{
                u32 val;
                if (__get_user(val, (uint32_t __user *) arg))
                    return -EFAULT;

                dprintk(1, "MVPCI_IOC_WRITE_MAXPAYLOAD: val=%d\n", val);
                pcie_set_mps(mvpci_drv->pci, val);
                mvpci_drv->maxpayloadsize = val;
                break;
            }
        case MVPCI_IOC_READ_FPGA_FAMILY:{
                dprintk(1, "device mvpci-%d: MVPCI_READ_FPGA_FAMILY: %d\n", mvpci_drv->nr, mvpci_drv->revision);
                /* only the lower 8 bits are significant, the others should be set to 0 */
                arg32 = (u32) mvpci_drv->revision;
                result = __put_user(arg32, (uint32_t __user *) arg);
                break;
            }
        case MVPCI_IOC_READ_FPGA_VERSION:{
                dprintk(1, "device mvpci-%d: MVPCI_READ_FPGA_VERSION: %d\n", mvpci_drv->nr, mvpci_drv->subsystem_id);
                /* only the lower 16 bits are significant, the others should be set to 0 */
                arg32 = (u32) mvpci_drv->subsystem_id;
                result = __put_user(arg32, (uint32_t __user *) arg);
                break;
            }
        case MVPCI_IOC_READ_IOWSEQ_BASE_ADDR:{
#ifdef USE_IOW_SEQUENCER
                dprintk(1,
                        "device mvpci-%d: MVPCI_IOC_READ_IOWSEQ_BASE_ADDR: %pad\n",
                        mvpci_drv->nr, &mvpci_drv->dma_iow_sequencer);
                result = copy_to_user((u64 __user *) arg, &mvpci_drv->dma_iow_sequencer, sizeof(u64));
#else
                dprintk(1,
                        "device mvpci-%d: MVPCI_IOC_READ_IOWSEQ_BASE_ADDR: NOT IMPLEMENTED\n", mvpci_drv->nr);
                result = 0;
#endif // USE_IOW_SEQUENCER
                break;
            }
        case MVPCI_IOC_READ_PCIE_LINK_WIDTH:{
                arg32 = get_pcie_link_width(mvpci_drv);
                dprintk(1, "device mvpci-%d: MVPCI_IOC_READ_PCIE_WIDTH: %d\n", mvpci_drv->nr, arg32);
                result = __put_user(arg32, (uint32_t __user *) arg);
                break;
            }
        case MVPCI_IOC_READ_PCIE_BUS_SPEED:{
                arg32 = get_pcie_link_speed(mvpci_drv);
                dprintk(1, "device mvpci-%d: MVPCI_IOC_READ_PCIE_BUS_SPEED: %d\n", mvpci_drv->nr, arg32);
                result = __put_user(arg32, (uint32_t __user *) arg);
                break;
            }
        case MVPCI_IOC_READ_PCIE_BANDWIDTH:{
                arg32 = get_pcie_link_bandwidth(mvpci_drv);
                dprintk(1, "device mvpci-%d: MVPCI_IOC_READ_PCIE_BANDWIDTH: %d\n", mvpci_drv->nr, arg32);
                result = __put_user(arg32, (uint32_t __user *) arg);
                break;
            }
        case MVPCI_IOC_READ_PCIE_DEVID:{
                struct pci_dev* pci = mvpci_drv->pci;
                arg32 = (pci->bus->number << 8) | pci->devfn;
                dprintk(1, "device mvpci-%d: MVPCI_IOC_READ_PCIE_DEVID: %d\n", mvpci_drv->nr, arg32);
                result = __put_user(arg32, (uint32_t __user *) arg);
                break;
            }
        default:{
                result = mvdma_ioctl(filp, cmd, arg);   /* use the dma "device's" ioctl */
                break;
            }
        }
    }
    return result;
}

static const struct file_operations mvpci_fops = {
    .owner = THIS_MODULE,
    .open = mvpci_open,
    .release = mvpci_release,
    .write = mvpci_write,
    .read = mvpci_read,
    .unlocked_ioctl = mvpci_ioctl,
};

static int camera_reset_write(int camera_num)
{
    int retval = -1;
    if (camera_num < number_of_devices) {
        struct mvpci_driver *mvpci_drv = mvpci_devices[camera_num];

        if (mvpci_drv && (mvpci_drv->nr == camera_num)
            && (mvpci_drv->initialized != 0)) {
            struct pci_dev *our_device = mvpci_drv->pci;
            if (mvpci_drv->fh.opened) {
#ifdef COWARDLY_REFUSAL
                dprintk(0, "device mvpci-%d: cowardly refusing to reset a device that is being used!\n", mvpci_drv->nr);
                return retval;
#else
                dprintk(0,
                        "device mvpci-%d: device is open %d times, but we are going to ignore it!\n",
                        mvpci_drv->nr, mvpci_drv->fh->opened);
#endif // COWARDLY_REFUSAL
            }
            if (our_device) {
                struct pci_bus *our_bus = our_device->bus;
                struct pci_dev *our_parent = our_bus->self;
                struct pci_bus *our_parents_bus = our_parent->bus;
                bool officially_removable = false;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0))
                /* crude TB detection: the device may be connected using thunderbolt (if the kernel is new enough)
                   so check if it can officially be unplugged i.e. if this is very probably a TB connection
                */
                if (dev_removable_is_valid(&our_device->dev)) {
                    officially_removable = dev_is_removable(&our_device->dev);
                }
#endif
                dprintk(0,
                        "device mvpci-%d: our camera device (0x%04x:0x%04x) at %s is subordinate to the bridge (0x%04x:0x%04x) at %s\n",
                        mvpci_drv->nr, our_device->vendor,
                        our_device->device,
                        pci_name(our_device), our_parent->vendor, our_parent->device, pci_name(our_parent));
                if (officially_removable) {
                    dprintk(0, "camera device is officially removable\n");
                }

                /* temporarily disable AER reporting from the bridge */
                pci_disable_pcie_error_reporting(our_parent);

                /* remove the device and signal to the km that it has gone
                 * this will make mvpci_drv invalid, so do not use beyond this point!
                 */
                pci_stop_and_remove_bus_device(our_device);
                mdelay(3000);

                /* re-enable AER reporting from the bridge */
                pci_enable_pcie_error_reporting(our_parent);

                /* TB connections do not like us resetting the parent bus device and do not need this either, so skip in that case */
                if (!officially_removable) {
                    dprintk(0, "resetting the secondary bus...\n");
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 18, 0)) || defined(RHEL_MAJOR) && defined(RHEL_MINOR) && ((RHEL_MAJOR == 8) && (RHEL_MINOR == 4))
                    /* CentOS etc. uses the new function name 'pci_bridge_secondary_bus_reset' in kernel 4.18.0
                     * even though the name wasn't officially changed until 4.19.0 
                     */
                    pci_bridge_secondary_bus_reset(our_parent); /* includes a 1 sec wait */
#else
                    pci_reset_bridge_secondary_bus(our_parent); /* includes a 1 sec wait */
#endif
                }
                dprintk(0, "rescanning the secondary bus...\n");
                pci_rescan_bus(our_parents_bus);
                mdelay(3000);
                retval = 0;
            }
        }
    }
    return retval;
}

static ssize_t camera_reset_write_all(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count)
{
    int nr;
    /* camera_reset0, camera_reset1, ... */
    /* 0123456789012, 0123456789012, ... */
    const char *name = (char *)attr->attr.name + 12;    /* offset to the numerical part */

    /* extract the device number from the name of the sysfs pseudo name */
    if ((kstrtoint(name, 10, &nr) == 0) && (nr >= 0) && (nr < number_of_devices)) {
        dprintk(0, "Calling camera_reset_write(%d)\n", nr);
        camera_reset_write(nr);
    }
    return count;
}

static enum hrtimer_restart hrtimer_handler(struct hrtimer *timer)
{
    dprintk(0, "READ TIMER expired.\n");
    return HRTIMER_NORESTART;
}

static int mvpci_probe(struct pci_dev *pci_dev, const struct pci_device_id *pci_id)
{
    int i;
    int result;
    unsigned char lat;
    struct mvpci_driver *mvpci_drv = 0;
    int num_vectors = 0;
    int mvpci_num = 0;

    if (number_of_devices >= num_cameras) {
        pr_err("%s: ERROR! Trying to use more than %d cameras. Maybe you need to change the 'num_cameras' parameter?\n",
               __FUNCTION__, num_cameras);
        return -EIO;
    }

    for (i = 0; i < num_cameras; i++) {
        mvpci_drv = mvpci_devices[i];
        if (!mvpci_drv) {
            dprintk(1, "%s: create new device %d\n", __FUNCTION__, i);
            mvpci_drv = (struct mvpci_driver *)kzalloc(sizeof(struct mvpci_driver), GFP_KERNEL);
            if (!mvpci_drv)
                return -ENOMEM;
            mvpci_devices[i] = mvpci_drv;
            mvpci_num = i;
            break;
        }
        /* find the first free device structure - we might be reinitialising after a PCI rescan */
        else if (mvpci_drv->initialized == 0) {
            mvpci_num = i;
            break;
        }
    }

    dprintk(1, "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
    dprintk(0, "MVPCI PCI device %d found.\n", mvpci_num);

    /* clear everything! */
    memset(mvpci_drv, 0, sizeof(*mvpci_drv));

    mutex_init(&mvpci_drv->dev_lock);
    mvpci_drv->nr = mvpci_num;
    mvpci_drv->pci = pci_dev;
    mvpci_drv->id = pci_dev->device;
    mvpci_drv->flags = 0;
    mvpci_drv->initialized = 0;
    mvpci_drv->fh.mvpci_drv = mvpci_drv;
    sprintf(mvpci_drv->name, "mvpci-%d", mvpci_drv->nr);

#if defined(CONFIG_NUMA) && defined(USE_NUMA)
    /* try to get the node number for this device */
    mvpci_drv->node = pcibus_to_node(mvpci_drv->pci->bus);
    if (mvpci_drv->node < first_node(node_online_map))
        mvpci_drv->node = first_node(node_online_map);
#endif /* CONFIG_NUMA */

    mvpci_drv->devnum = MKDEV(MAJOR(mvpci_devnum), MINOR(mvpci_devnum) + mvpci_drv->nr);
    dprintk(0, "MAJOR=%d, minor=%d\n", MAJOR(mvpci_devnum), MINOR(mvpci_devnum) + mvpci_drv->nr);

    if (pci_enable_device(mvpci_drv->pci)) {
        dprintk(0, "can't enable device %d\n", mvpci_drv->nr);
        mvpci_drv->nr = -1;
        return -EIO;
    }
    pci_set_master(mvpci_drv->pci);
    pci_set_drvdata(mvpci_drv->pci, mvpci_drv);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 7, 0)) || \
    (defined(RHEL_MAJOR) && defined(RHEL_MINOR) && ((RHEL_MAJOR == 8) && (RHEL_MINOR == 4))) || \
    (defined(CONFIG_SUSE_KERNEL) && (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 3, 18)))
    /* CentOS etc. uses the new function name 'pci_aer_clear_nonfatal_status' in kernel 4.18.0
     * even though the name wasn't officially changed until 5.7.0 .
     * SuSE uses the new function in its 5.3.18 kernel too!
     */
    pci_aer_clear_nonfatal_status(mvpci_drv->pci);
#else
    pci_cleanup_aer_uncorrect_error_status(mvpci_drv->pci);
#endif

    if (!pci_set_dma_mask(mvpci_drv->pci, DMA_BIT_MASK(64))) {
        dprintk(1, "64 bit DMA available on PCIe device %s.\n", mvpci_drv->name);
    } else {
        dprintk(0, "Error! 64 bit DMA NOT available on PCIe device %s.\n", mvpci_drv->name);
        mvpci_drv->nr = -1;
        return -EIO;
    }

#ifdef USE_COHERENT_DMA
    if (!dma_set_coherent_mask(&mvpci_drv->pci->dev, DMA_BIT_MASK(64))) {
        dprintk(1, "Coherent DMA is available on PCIe device %s.\n", mvpci_drv->name);
        mvpci_drv->coherent = 1;
    } else {
        dprintk(0, "Warning! Coherent DMA is NOT available on PCIe device %s.\n", mvpci_drv->name);
    }
#endif

    if (pci_resource_flags(mvpci_drv->pci, 0) & IORESOURCE_MEM) {
        dprintk(0, "found MMIO @ BAR-0 of length 0x%llx on device %d\n",
                pci_resource_len(mvpci_drv->pci, 0), mvpci_drv->nr);
    } else {
        dprintk(0, "no MMIO available on device %d\n", mvpci_drv->nr);
        mvpci_drv->nr = -1;
        return -EIO;
    }

    if (pci_request_regions(mvpci_drv->pci, mvpci_drv->name)) {
        dprintk(0, "can't request iomem for device %d\n", mvpci_drv->nr);
        mvpci_drv->nr = -1;
        return -EBUSY;
    } else {
        mvpci_drv->flags |= MVPCI_FLAG_RES0;
    }

    mvpci_drv->maxpayloadsize = read_maxpayload(mvpci_drv->pci);
    mvpci_drv->maxreadreq = read_maxreadreq(mvpci_drv->pci);
    dprintk(0, "%s: max. packet size: %d\n", __FUNCTION__, pcie_get_mps(mvpci_drv->pci));

    pci_read_config_byte(mvpci_drv->pci, PCI_REVISION_ID, &mvpci_drv->revision);
    pci_read_config_word(mvpci_drv->pci, PCI_SUBSYSTEM_ID, &mvpci_drv->subsystem_id);
    dprintk(0,
            "device 0x%04x (revision: 0x%x, subsystem_id: 0x%04x) at %s\n",
            mvpci_drv->id, mvpci_drv->revision, mvpci_drv->subsystem_id, pci_name(mvpci_drv->pci));

    dprintk(0, "link bandwidth: %d Mb/s (width = %d, bus speed = %d)\n",
            get_pcie_link_bandwidth(mvpci_drv), get_pcie_link_width(mvpci_drv), get_pcie_link_speed(mvpci_drv));

    pci_read_config_byte(mvpci_drv->pci, PCI_LATENCY_TIMER, &lat);
    dprintk(1, "non msi irq: %d, latency: %d, resource_start: 0x%llx\n",
            mvpci_drv->pci->irq, lat, (unsigned long long)pci_resource_start(mvpci_drv->pci, 0));

    mvpci_drv->mmio = pci_iomap(mvpci_drv->pci, 0, pci_resource_len(mvpci_drv->pci, 0));
    if (mvpci_drv->mmio == NULL) {
        pr_err("%s: ERROR: device %d pci_iomap failed\n", __FUNCTION__, mvpci_drv->nr);
        result = -EIO;
        goto fail;
    }
    dprintk(1, "mapped %d kB register space to 0x%px\n",
            (u32) (pci_resource_len(mvpci_drv->pci, 0) >> 10), mvpci_drv->mmio);

#ifdef USE_IOW_SEQUENCER
    //mvpci_drv->iow_sequencer = dma_alloc_coherent(&mvpci_drv->pci->dev, MVPCI_IOW_SEQUENCER_SIZE, &mvpci_drv->dma_iow_sequencer, GFP_DMA);
    mvpci_drv->iow_sequencer = dma_alloc_coherent(&mvpci_drv->pci->dev, MVPCI_IOW_SEQUENCER_SIZE, &mvpci_drv->dma_iow_sequencer, GFP_KERNEL);
    if (!mvpci_drv->iow_sequencer) {
        pr_err("%s: ERROR: device %d reserving io_writer sequencer memory failed\n", __FUNCTION__, mvpci_drv->nr);
        result = -EIO;
        goto fail;
    }
    dprintk(1,
            "reserved %lld kB for the io_writer sequencer @ 0x%pK/%pad\n",
            MVPCI_IOW_SEQUENCER_SIZE / _kB, mvpci_drv->iow_sequencer, &mvpci_drv->dma_iow_sequencer);
#endif // USE_IOW_SEQUENCER

    for (i = 0; i < MVPCI_MSI_MAX_VECTORS; i++) {
        mvpci_drv->msi_x_entry[i].entry = i;
    }

    /* disable Interrupts from the hardware */
    mvwrite(0x00, MVPCI_IO_CONTROL);

#if defined(TRY_MSIX) || defined(TRY_MSI)
    /* try to use MSI-X or MSI interrupts */
    dprintk(0, "trying to use %d vector(s) on device mvpci-%d\n", MVPCI_MSI_MAX_VECTORS, mvpci_num);
#ifdef TRY_MSIX
    num_vectors = pci_msix_vec_count(mvpci_drv->pci);
    if (num_vectors > 0) {
        dprintk(0, "support for %d MSI-X vectors on device mvpci-%d\n", num_vectors, mvpci_num);
    }
#endif /* TRY_MSIX */
#ifdef TRY_MSI
    num_vectors = pci_msi_vec_count(mvpci_drv->pci);
    if (num_vectors > 0) {
        dprintk(0, "support for %d MSI vectors on device mvpci-%d\n", num_vectors, mvpci_num);
    }
#endif /* TRY_MSI */
#endif //defined(TRY_MSIX) || defined(TRY_MSI)

#if defined(TRY_MSIX)
    result =
        pci_alloc_irq_vectors(mvpci_drv->pci, MVPCI_MSI_MIN_VECTORS,
                              MVPCI_MSI_MAX_VECTORS, PCI_IRQ_MSI | PCI_IRQ_MSIX | PCI_IRQ_AFFINITY);
#elif defined(TRY_MSI)
    result =
        pci_alloc_irq_vectors(mvpci_drv->pci, MVPCI_MSI_MIN_VECTORS,
                              MVPCI_MSI_MAX_VECTORS, PCI_IRQ_MSI | PCI_IRQ_AFFINITY);
#else
    result = -1;
#endif /* TRY_MSI */

    if (result < 0) {
        /* only LEGACY INTERRUPTS will work */
        result = pci_alloc_irq_vectors(mvpci_drv->pci, 1, 1, PCI_IRQ_LEGACY);
        if (result < 0) {
            pr_err
                ("%s: ERROR: device mvpci-%d: failed to obtain a single legacy interrupt for MSeq!\n",
                 __FUNCTION__, mvpci_num);
            goto fail;
        }
        num_vectors = result;
        dprintk(0, "device mvpci-%d: legacy interrupts only will be used on this system.\n", mvpci_num);
        mvpci_drv->flags |= MVPCI_FLAG_IRQ;
    } else {
        /* MSI and MSI-X return 1 or more vectors */
        num_vectors = result;
        dprintk(0,
                "device mvpci-%d: %d MSI or MSI-X interrupt vectors supplied on this system.\n",
                mvpci_num, num_vectors);
        if ((num_vectors != MVPCI_MSI_MIN_VECTORS)
            && (num_vectors != MVPCI_MSI_MAX_VECTORS)) {
            pr_err
                ("%s: ERROR: device mvpci-%d: system supplies %d interrupts, but the driver needs either 1 or %d!\n",
                 __FUNCTION__, mvpci_num, result, MVPCI_MSI_MAX_VECTORS);
            goto fail;
        }
        mvpci_drv->flags |= MVPCI_FLAG_MSI;
    }

    for (i = 0; i < MVPCI_MSI_MAX_VECTORS; i++) {
        irq_handler_t handler = NULL;
        int uio_interrupt = 0;
        char uio_dev_name[10] = "";

        result = -1;
        switch (i) {
        case INT_MSEQ:
            if (num_vectors == 1)
                /* legacy interrupt  or just 1 MSI available */
                handler = mvpci_isr_single;
            else
                /* >1 MSI or MS-X interrupt */
                handler = mvpci_isr_mseq;
            break;
        case INT_IO:
            uio_interrupt = 1;
            strcpy(uio_dev_name, "mvio");
            handler = (irq_handler_t) MSI_UIO_ISR(INT_IO);
            break;
        case INT_STAT:
            uio_interrupt = 1;
            strcpy(uio_dev_name, "mvstat");
            handler = (irq_handler_t) MSI_UIO_ISR(INT_STAT);
            break;
        case INT_SC:
            uio_interrupt = 1;
            strcpy(uio_dev_name, "mvsc");
            handler = (irq_handler_t) MSI_UIO_ISR(INT_SC);
            break;
        default:
            break;
        }

        if (uio_interrupt) {
            const char *name;

            /* register a UIO interrupt handler */
            result = set_allocate_name(&name, "%s:mvpci%d", uio_dev_name, mvpci_num);
            if (result) {
                pr_err
                    ("%s: ERROR: can't set uio interrupt name for device %d, vector %d, IRQ %d\n",
                     __FUNCTION__, mvpci_num, i, mvpci_drv->msi_x_entry[i].vector);
                goto fail;
            }

            mvpci_drv->uio_info[i].name = name;
            mvpci_drv->uio_info[i].version = MVPCI_VERSION_STR;
            if (num_vectors == 1) {
                mvpci_drv->msi_x_entry[i].vector = UIO_IRQ_CUSTOM;
                mvpci_drv->uio_info[i].irq = UIO_IRQ_CUSTOM;
                mvpci_drv->uio_info[i].priv = (void *)mvpci_drv;
                dprintk(0, "device mvpci-%d: set uio interrupt name %s, vector %d, *CUSTOM* IRQ\n", mvpci_num, name, i);
            } else {
                mvpci_drv->msi_x_entry[i].vector = pci_irq_vector(mvpci_drv->pci, i);
                mvpci_drv->uio_info[i].irq = mvpci_drv->msi_x_entry[i].vector;
                mvpci_drv->uio_info[i].irq_flags = IRQF_SHARED;
                mvpci_drv->uio_info[i].handler = (irq_uio_handler_t) handler;
                mvpci_drv->uio_info[i].priv = (void *)mvpci_drv;
                dprintk(0,
                        "device mvpci-%d: set uio interrupt name %s, vector %d, IRQ %d\n",
                        mvpci_num, name, i, mvpci_drv->msi_x_entry[i].vector);
            }

            result = uio_register_device(&pci_dev->dev, &mvpci_drv->uio_info[i]);

            if (result < 0) {
                pr_err
                    ("%s: ERROR: device mvpci-%d: uio_register_device failed using vector %d, IRQ %d, returns %d\n",
                     __FUNCTION__, mvpci_num, i, mvpci_drv->msi_x_entry[i].vector, result);
                /* tidy up */
                mvpci_drv->msi_x_entry[i].vector = 0;
                memset(&mvpci_drv->uio_info[i], 0, sizeof(struct uio_info));
                kfree_const(name);
                goto fail;
            }
            /* else OK */

        } else if (handler) {
            /* register a regular interrupt handler */
            mvpci_drv->msi_x_entry[i].vector = pci_irq_vector(mvpci_drv->pci, i);
            if ((i == INT_MSEQ)
                && mvpci_drv->flags & MVPCI_FLAG_IRQ) {
                dprintk(0,
                        "device mvpci-%d trying to use a single legacy interrupt %d for MSeq.\n",
                        mvpci_num, mvpci_drv->msi_x_entry[INT_MSEQ].vector);
                mvpci_drv->pci->irq = mvpci_drv->msi_x_entry[INT_MSEQ].vector;
            } else {
                dprintk(0,
                        "device mvpci-%d trying to use a MSI interrupt %d for MSeq.\n",
                        mvpci_num, mvpci_drv->msi_x_entry[i].vector);
            }

            result = request_irq(mvpci_drv->msi_x_entry[i].vector,
                                 handler, IRQF_SHARED, mvpci_drv->name, (void *)mvpci_drv);
        } else {
            /* unused interrupt - ignore */
            continue;
        }

        if (result < 0) {
            pr_err
                ("%s: ERROR: device mvpci-%d: vector %d, IRQ %d\n",
                 __FUNCTION__, mvpci_num, i, mvpci_drv->msi_x_entry[i].vector);
            goto fail;
        } else {
            dprintk(0, "device mvpci-%d: using vector %d, IRQ %d\n", mvpci_num, i, mvpci_drv->msi_x_entry[i].vector);
        }
    }                           /* for */

    /* setup character device */
    cdev_init(&mvpci_drv->cdev, &mvpci_fops);
    if ((result = cdev_add(&mvpci_drv->cdev, mvpci_drv->devnum, 1))) {
        pr_err("%s: ERROR: device %d cdev_add failed %d\n", __FUNCTION__, mvpci_drv->nr, result);
        goto fail;
    } else {
        dprintk(1, "device %d cdev_add succeeded\n", mvpci_drv->nr);
    }

    /* setup dma 'device' */
    result = mvdma_probe(mvpci_drv);
    if (result < 0) {
        pr_err("%s: ERROR: can't setup DMA structures\n", __FUNCTION__);
        goto fail;
    } else {
        dprintk(1, "Setup DMA structures OK\n");
    }

    /* read timeout timer */
    hrtimer_init(&mvpci_drv->read_hrtimer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
    mvpci_drv->read_hrtimer.function = hrtimer_handler;

    /* disable interrupts on the CPU */
    mvinterrupts_enable_all(mvpci_drv, 0);
#ifdef USE_PCI_ERROR_HANDLERS
    /* enable AER reporting */
    pci_enable_pcie_error_reporting(mvpci_drv->pci);
#endif //  USE_PCI_ERROR_HANDLERS
    number_of_devices++;
    mvpci_drv->initialized = 1;
    dprintk(0, "initialised mvpci device %d (total devices = %d)\n", mvpci_drv->nr, number_of_devices);
    return 0;

fail:
#ifdef USE_IOW_SEQUENCER
    if (mvpci_drv->iow_sequencer) {
        dprintk(1, "device mvpci-%d: free io_writer sequencer memory\n", mvpci_drv->nr);
        dma_free_coherent(&mvpci_drv->pci->dev,
                          MVPCI_IOW_SEQUENCER_SIZE, mvpci_drv->iow_sequencer, mvpci_drv->dma_iow_sequencer);
        mvpci_drv->iow_sequencer = NULL;
        mvpci_drv->dma_iow_sequencer = 0;
    }
#endif // USE_IOW_SEQUENCER
    mvpci_free_pci(mvpci_drv);
    return result;
}

static void mvpci_remove(struct pci_dev *pci_dev)
{
    struct mvpci_driver *mvpci_drv = pci_get_drvdata(pci_dev);

    dprintk(1, "removing device %d, name = %s, dev = %s...\n", mvpci_drv->nr, mvpci_drv->name, pci_name(pci_dev));
    mvpci_free_pci(mvpci_drv);
    if (number_of_devices > 0)
        number_of_devices--;
    dprintk(0, "device %d removed, %d devices remain\n", mvpci_drv->nr, number_of_devices);
    return;
}

#ifdef USE_PCI_ERROR_HANDLERS
static pci_ers_result_t mvpci_error_detected(struct pci_dev *pci_dev, pci_channel_state_t state)
{
    return PCI_ERS_RESULT_CAN_RECOVER;
}

static pci_ers_result_t mvpci_slot_reset(struct pci_dev *pci_dev)
{
    return PCI_ERS_RESULT_RECOVERED;
}

static const struct pci_error_handlers mvpci_err_handler = {
    .error_detected = mvpci_error_detected,
    .slot_reset = mvpci_slot_reset,
};
#endif /* USE_PCI_ERROR_HANDLERS */

#ifndef PCI_VENDOR_ID_MV
#define PCI_VENDOR_ID_MV 0x4d56 /* MV */
#endif
#define PCI_ID_PRODUCT1 0x7014
#define PCI_ID_PRODUCT2 0x7024
#define PCI_ID_PRODUCT3 0x7012
#define PCI_ID_PRODUCT4 0x7022
static struct pci_device_id mvpci_pci_tbl[] = {
    {PCI_VENDOR_ID_MV, PCI_ID_PRODUCT1, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
    {PCI_VENDOR_ID_MV, PCI_ID_PRODUCT2, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
    {PCI_VENDOR_ID_MV, PCI_ID_PRODUCT3, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
    {PCI_VENDOR_ID_MV, PCI_ID_PRODUCT4, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
    {
     0,
     }
};

MODULE_DEVICE_TABLE(pci, mvpci_pci_tbl);

static struct pci_driver mvpci_pci_driver = {
    .name = "mvpci",
    .id_table = mvpci_pci_tbl,
    .probe = mvpci_probe,
    .remove = mvpci_remove,
#ifdef USE_PCI_ERROR_HANDLERS
    .err_handler = &mvpci_err_handler,
#endif /* USE_PCI_ERROR_HANDLERS */
};

static int mvpci_init_module(void)
{
#ifndef CONFIG_PCI_MSI
    pr_err("%s: ERROR! Kernel needs to be configured with CONFIG_PCI_MSI!\n", __FUNCTION__);
    return -EFAULT;
#endif /* CONFIG_PCI_MSI */

    mvpci_devices = (struct mvpci_driver **)kzalloc(num_cameras * sizeof(struct mvpci_driver *), GFP_KERNEL);
    if (!mvpci_devices) {
        pr_err("%s: ERROR! Insufficient memory to allocate mvpci_devices\n", __FUNCTION__);
        return -ENOMEM;
    }

    mvpci_sysfs_entries = (struct mvpci_sysfs **)kzalloc(num_cameras * sizeof(struct mvpci_sysfs *), GFP_KERNEL);
    if (!mvpci_sysfs_entries) {
        pr_err("%s: ERROR! Insufficient memory to allocate mvpci_sysfs_entries\n", __FUNCTION__);
        return -ENOMEM;
    }

    if (alloc_chrdev_region(&mvpci_devnum, 0, num_cameras, "mvpci") == 0) {
        dprintk(1, "Successfully registered mvpci as a character device\n");
    } else {
        pr_err("%s: ERROR! Can't register mvpci as a character device\n", __FUNCTION__);
        return -EFAULT;
    }

    dprintk(1, "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
    dprintk(0, "mvpci v%s loaded (major device: %d, debug level = 0x%x)\n",
            MVPCI_VERSION_STR, MAJOR(mvpci_devnum), debug);
    dprintk(0, "max. %d cameras may be connected. Change using module parameter: num_cameras\n", num_cameras);
    dprintk(1, "PAGE_SIZE on this system is 0x%08lx\n", PAGE_SIZE);

    /* our sysfs entries go here */
    mvpci_kobj = kobject_create_and_add("matrix-vision", kernel_kobj);
    if (!mvpci_kobj)
        return -ENOMEM;

    if (pci_register_driver(&mvpci_pci_driver) == 0) {
        dprintk(0, "Successfully registered mvpci as a PCI driver\n");
    } else {
        pr_err("%s: ERROR! Can't register mvpci as a PCI driver\n", __FUNCTION__);
        return -EFAULT;
    }
    {
        // create sysfs entries for devices that we have found
        unsigned nr;
        for (nr = 0; nr < number_of_devices; nr++) {
            struct mvpci_sysfs *sfs = (struct mvpci_sysfs *)kzalloc(sizeof(struct mvpci_sysfs), GFP_KERNEL);
            const char *name;
            int result = set_allocate_name(&name, "camera_reset%d", nr);
            if (!result) {
                sfs->camera_reset_attribute.attr.name = name;
                sfs->camera_reset_attribute.attr.mode = 0220;   /* write only function */
                sfs->camera_reset_attribute.show = NULL;    /* no read function */
                sfs->camera_reset_attribute.store = camera_reset_write_all; /* write function */
                sfs->camera_attrs[0] = &sfs->camera_reset_attribute.attr;
                sfs->camera_attrs[1] = NULL;    /* just a single entry, mark the end */
                sfs->camera_attr_group.attrs = sfs->camera_attrs;
                result = sysfs_create_group(mvpci_kobj, &sfs->camera_attr_group);
            }
            if (result) {
                sfs->camera_reset_attribute.attr.name = NULL;
                pr_err("%s: ERROR! Can't register 'camera_reset%d' with sysfs. Device reset is NOT possible.\n",
                       __FUNCTION__, nr);
            } else {
                dprintk(0, "Registered 'camera_reset%d' with sysfs. Device reset is possible.\n", nr);
            }
            mvpci_sysfs_entries[nr] = sfs;
        }
    }
    return 0;
}

static void mvpci_cleanup_module(void)
{
    unsigned nr;

    for (nr = 0; nr < number_of_devices; nr++) {
        struct mvpci_driver *mvpci_drv = mvpci_devices[nr];
        if (mvpci_drv && mvpci_drv->nr >= 0) {
            struct mvpci_sysfs *sfs = mvpci_sysfs_entries[nr];
            if (sfs) {
                /* remove sysfs entries for devices that exist at the moment */
                dprintk(1, "Free sysfs entry %d\n", nr);
                if (sfs->camera_reset_attribute.attr.name) {
                    sysfs_remove_group(&(mvpci_drv->pci->dev.kobj), &sfs->camera_attr_group);
                    /* we created this string using 'set_allocate_name' */
                    kfree_const(sfs->camera_reset_attribute.attr.name);
                }
                kfree(mvpci_sysfs_entries[nr]);
                mvpci_sysfs_entries[nr] = NULL;
            }
            /* remove character devices */
            cdev_del(&mvpci_drv->cdev);
        }
    }
    /* remove our sysfs directory */
    if (mvpci_kobj)
        kobject_put(mvpci_kobj);

    unregister_chrdev_region(mvpci_devnum, num_cameras);
    pci_unregister_driver(&mvpci_pci_driver);
    dprintk(0, "Unregistered mvPCI driver\n");

    for (nr = 0; nr < num_cameras; nr++) {
        if (mvpci_devices[nr]) {
            dprintk(1, "Free driver %d\n", nr);
            kfree(mvpci_devices[nr]);
            mvpci_devices[nr] = NULL;
        }
    }
    kfree(mvpci_devices);
    return;
}

module_init(mvpci_init_module);
module_exit(mvpci_cleanup_module);
